class TestReadme:
    def test_instancing_env_detailed(self):
        from tradingenv import TradingEnv
        from tradingenv.contracts import ETF
        from tradingenv.spaces import BoxPortfolio
        from tradingenv.state import IState
        from tradingenv.rewards import RewardLogReturn
        from tradingenv.broker.fees import BrokerFees
        import yfinance

        # Load close prices from Yahoo Finance and specify contract types.
        prices = yfinance.Tickers(['SPY', 'TLT', 'TBIL']).history(period="12mo", progress=False)['Close'].tz_localize(None)
        prices.columns = [ETF('SPY'), ETF('TLT'), ETF('TBIL')]

        # Instance the trading environment.
        env = TradingEnv(
            action_space=BoxPortfolio([ETF('SPY'), ETF('TLT')], low=-1, high=+1, as_weights=True),
            state=IState(),
            reward=RewardLogReturn(),
            prices=prices,
            initial_cash=1_000_000,
            latency=0,  # seconds
            steps_delay=1,  # trades are implemented with a delay on one step
            broker_fees=BrokerFees(
                markup=0.005,  # 0.5% broker markup on deposit rate
                proportional=0.0001,  # 0.01% fee of traded notional
                fixed=1,  # $1 per trade
            ),
        )

        # OpenAI/gym protocol. Run an episode in the environment.
        # env can be passed to RL agents of ray/rllib or stable-baselines3.
        obs = env.reset()
        done = False
        while not done:
            action = env.action_space.sample()
            obs, reward, done, info = env.step(action)
        return env, prices

    def test_backtest_60_40_old(self):
        env, prices = self.test_instancing_env_detailed()
        # BEGIN OMIT
        from tradingenv.policy import AbstractPolicy

        class Portfolio6040(AbstractPolicy):
            """Implement logic of your investment strategy or RL agent here."""

            def act(self, state):
                """Invest 60% of the portfolio in SPY ETF and 40% in TLT ETF."""
                return [0.6, 0.4]

        # Run the backtest.
        track_record = env.backtest(
            policy=Portfolio6040(),
            risk_free=prices['TBIL'],
            benchmark=prices['SPY'],
        )

        # The track_record object stores the results of your backtest.
        track_record.tearsheet()
        track_record.fig_net_liquidation_value()
        # END OMIT

    def test_instancing_env_lazy(self):
        # BEGIN OMIT
        from tradingenv.env import TradingEnvXY
        import yfinance

        # Load data from Yahoo Finance.
        tickers = yfinance.Tickers(['SPY', 'TLT', 'TBIL', '^IRX'])
        data = tickers.history(period="12mo", progress=False)['Close'].tz_localize(None)
        Y = data[['SPY', 'TLT']]
        X = Y.rolling(12).mean() - Y.rolling(26).mean()

        # Default instance of the trading environment.
        env = TradingEnvXY(X, Y)

        # Run an episode in the environment.
        obs = env.reset()
        done = False
        while not done:
            action = env.action_space.sample()
            obs, reward, done, info = env.step(action)
        # END OMIT
        return env, data, X, Y

    def test_instancing_env_custom(self):
        from tradingenv.env import TradingEnvXY
        env, data, X, Y = self.test_instancing_env_lazy()

        # BEGIN OMIT
        env = TradingEnvXY(
            X=X,                      # Use moving averages crossover as features
            Y=Y,                      # to trade SPY and TLT ETFs.
            transformer='z-score',    # Features are standardised to N(0, 1).
            reward='logret',          # Reward is the log return of the portfolio at each step,
            cash=1000000,             # starting with $1M.
            spread=0.0002,            # Transaction costs include a 0.02% spread,
            markup=0.005,             # a 0.5% broker markup on deposit rate,
            fee=0.0002,               # a 0.02% dealing fee of traded notional
            fixed=1,                  # and a $1 fixed fee per trade.
            margin=0.02,              # Do not trade if trade size is smaller than 2% of the portfolio.
            rate=data['^IRX'] / 100,  # Rate used to compute the yield on idle cash and cost of leverage.
            latency=0,                # Trades are implemented with no latency
            steps_delay=1,            # but a delay of one day.
            window=1,                 # The observation is the current state of the market,
            clip=5.,                  # clipped between -5 and +5 standard deviations.
            max_long=1.5,             # The maximum long position is 150% of the portfolio,
            max_short=-1.,            # the maximum short position is 100% of the portfolio.
            calendar='NYSE',          # Use the NYSE calendar to schedule trading days.
        )
        # END OMIT
        return env, data

    def test_backtest_60_40(self):
        env, data, X, Y = self.test_instancing_env_lazy()

        # BEGIN OMIT
        from tradingenv.policy import AbstractPolicy

        class Portfolio6040(AbstractPolicy):
            """Implement logic of your investment strategy or RL agent here."""

            def act(self, state):
                """Invest 60% of the portfolio in SPY ETF and 40% in TLT ETF."""
                return [0.6, 0.4]

        # Run the backtest.
        track_record = env.backtest(
            policy=Portfolio6040(),
            risk_free=data['TBIL'],
            benchmark=data['SPY'],
        )

        # The track_record object stores the results of your backtest.
        track_record.tearsheet()
        # END OMIT


"""The reward is a parameter of TradingEnv. Valid rewards must implement the
abstract class AbstractReward."""
import tradingenv
from abc import ABC, abstractmethod
import numpy as np
from typing import Union


def make_reward(reward: Union["AbstractReward", str]):
    """Valid reward are implementation of the interface AbstractReward. However,
    this method allows the user to specify the reward as a string and the
    corresponding id will be retrieved from tradingenv.rewards."""
    if isinstance(reward, str):
        reward = getattr(tradingenv.rewards, reward)()
    if not isinstance(reward, AbstractReward):
        raise ValueError(
            "{} is an invalid reward. Valid rewards must be objects "
            "implementing tradingenv.rewards.AbstractReward or strings "
            "indicating referring to class _names defined in "
            "tradingenv.rewards."
        )
    return reward


class AbstractReward(ABC):
    """All custom rewards must implement this interface."""

    @abstractmethod
    def calculate(self, env: "tradingenv.env.TradingEnv") -> float:
        """Return float associated with the last reward of the agent,
        generally manipulating stuff from env.broker.track_record. See
        implementations for examples."""

    def reset(self) -> None:
        """Reset values of this class, if any."""


class RewardPnL(AbstractReward):
    """Profit and Loss reward."""

    def calculate(self, env: "tradingenv.env.TradingEnv") -> float:
        nlv_last_rebalancing = env.broker.track_record[-1].context_pre.nlv
        nlv_now = env.broker.net_liquidation_value()
        return float(nlv_now - nlv_last_rebalancing)


class RewardLogReturn(AbstractReward):
    """Log change of the net liquidation value of the account at each step."""

    def calculate(self, env: "tradingenv.env.TradingEnv") -> float:
        nlv_last_rebalancing = env.broker.track_record[-1].context_pre.nlv
        nlv_now = env.broker.net_liquidation_value()
        return float(np.log(nlv_now / nlv_last_rebalancing))


class LogReturn(AbstractReward):
    def __init__(self, scale: float = 1., clip: float = 2.):
        """
        Parameters
        ----------
        scale
            Reward is divided by this number before being returned. This is a
            helper to rescale the reward closer to a [-1, +1] range.
        clip
            Rewards larger than this clip value are truncated.

        Notes
        -----
        For a good reference on why reward shaping is important see [1].

        References
        ----------
        [1] van Hasselt, Hado P., et al. "Learning values across many orders of
        magnitude." Advances in neural information processing systems 29 (2016).
        """
        self.scale = scale
        self.clip = clip

    def calculate(self, env: "tradingenv.env.TradingEnv") -> float:
        nlv_last_rebalancing = env.broker.track_record[-1].context_pre.nlv
        nlv_now = env.broker.net_liquidation_value()
        ret = np.log(nlv_now / nlv_last_rebalancing)
        ret /= self.scale
        ret = np.clip(ret, -self.clip, +self.clip)
        return float(ret)


class RewardSimpleReturn(AbstractReward):
    """Simple change of the net liquidation value of the account at each
    step."""

    def calculate(self, env: "tradingenv.env.TradingEnv") -> float:
        nlv_last_rebalancing = env.broker.track_record[-1].context_pre.nlv
        nlv_now = env.broker.net_liquidation_value()
        return float(nlv_now / nlv_last_rebalancing) - 1


class RewardDifferentialSharpeRatio:
    """An elegant online Sharpe ratio which uses the second order Taylor
    expansion. I still wonder how problematic this approximation might be?

    References
    ----------
    http://papers.nips.cc/paper/1551-reinforcement-learning-for-trading.pdf
    https://quant.stackexchange.com/questions/37969/what-s-the-derivative-of-the-sharpe-ratio-for-one-asset-trying-to-optimize-on-i
    https://www.reddit.com/r/algotrading/comments/9xkvby/how_to_calculate_differential_sharpe_ratio/
    """

    def calculate(self, env: "tradingenv.env.TradingEnv") -> float:
        raise NotImplementedError()



from tradingenv.broker.allocation import NrContracts, Weights
from tradingenv.broker.rebalancing import Rebalancing
from tradingenv.broker.broker import Broker
from tradingenv.broker.trade import Trade
from tradingenv.events import EventNBBO, EventContractDiscontinued
from tradingenv.exchange import Exchange
from tradingenv.contracts import Cash, ETF, VX, ES, Rate, FutureChain
from collections import namedtuple
from datetime import datetime
import numpy as np
import pytest
Pair = namedtuple('Pair', ['allocation', 'expected'])
Imbalance = namedtuple('Pair', ['holdings', 'target', 'expected'])
MEASURES = ['weight', 'nr-contracts']
NOW = datetime.now()


def make_broker(deposit: float = 1000000, initial_holdings=None):
    initial_holdings = initial_holdings or dict()
    nbbos = [
        EventNBBO(datetime.now(), Cash(), 1, 1),
        EventNBBO(datetime.now(), Rate("FED funds rate"), 0.0182, 0.0182),
        EventNBBO(datetime.now(), ETF('SPY'), 293.24, 293.25),
        EventNBBO(datetime.now(), ETF('IEF'), 112.67, 112.68),
        EventNBBO(datetime.now(), VX(2020, 3), 19.20, 19.25),
        EventNBBO(datetime.now(), ES(2020, 3), 2941.75, 2942),
        EventNBBO(datetime.now(), ES(2020, 6), 2940, 2940.25),
        EventNBBO(datetime.now(), ES(2020, 9), 2938.75, 2939),
        EventNBBO(datetime.now(), ES(2020, 12), 2935.25, 2935),
    ]
    exchange = Exchange()
    for nbbo in nbbos:
        exchange.process_EventNBBO(nbbo)
    broker = Broker(exchange=exchange, base_currency=Cash(), deposit=deposit)
    for contract, quantity in initial_holdings.items():
        broker._holdings_quantity[contract] = quantity
    return broker


class TestRebalancing:
    def test_make_trades_does_not_raise_if_missing_not_needed_price(self):
        """Note that the price for ETF('IEF') if missing, but the _rebalancing
        proceeds anyway because the target weight is zero, so the price is not
        needed."""
        nlv = 100
        broker = make_broker(deposit=nlv)
        event = EventContractDiscontinued(datetime.now(), ETF('IEF'))
        broker.exchange[ETF('IEF')].terminate(event)
        assert np.isnan(broker.exchange[ETF('IEF')].mid_price)
        rebalancing = Rebalancing(
            contracts=[Cash(), ETF('SPY'), ETF('IEF')],
            allocation=[0.4, 0.6, 0.0],
            time=NOW,
        )
        actual = rebalancing.make_trades(broker)
        expected = [Trade(NOW, ETF('SPY'), nlv * 0.6 / 293.25, 293.24, 293.25)]
        assert actual == expected

    def test_make_trades_raise_if_missing_needed_price(self):
        """Note that the price for ETF('IEF') if missing, but the _rebalancing
        proceeds anyway because the target weight is zero, so the price is not
        needed."""
        nlv = 100
        broker = make_broker(deposit=nlv)
        event = EventContractDiscontinued(datetime.now(), ETF('IEF'))
        broker.exchange[ETF('IEF')].terminate(event)
        assert np.isnan(broker.exchange[ETF('IEF')].mid_price)
        rebalancing = Rebalancing(
            contracts=[Cash(), ETF('SPY'), ETF('IEF')],
            allocation=[0.4, 0.6, 0.2],
            time=NOW,
        )
        with pytest.raises(ValueError):
            rebalancing.make_trades(broker)

    def test_make_trades_when_absolute_is_false(self, mocker):
        """Absolute is True by default."""
        imbalance = {
            Cash(): 100,
            ETF('SPY'): -12.1,
            VX(2020, 3): 1.1,
            FutureChain(contracts=[ES(2020, 3), ES(2020, 6), ES(2020, 9)]): -1.2,
        }
        mocker.patch.object(
            target=Weights,
            attribute=Weights._to_nr_contracts.__name__,
            return_value=NrContracts(imbalance),
        )
        rebalancing = Rebalancing(absolute=False, time=NOW)
        broker = make_broker(deposit=100)
        actual = rebalancing.make_trades(broker)
        assert actual == [
            Trade(NOW, ETF('SPY'), -12.1, 293.24, 293.25),
            Trade(NOW, VX(2020, 3), 1.1, 19.20, 19.25),
            Trade(NOW, ES(2020, 3), -1.2, 2941.75, 2942),
        ]

    @pytest.mark.parametrize(
        argnames=['allocation', 'expected'],
        argvalues=[
            Pair(
                # to all cash.
                allocation={Cash(): 1, ETF('SPY'): 0, ETF('IEF'): 0},
                expected=[],
            ),
            Pair(
                # to balanced portfolio.
                allocation={Cash(): 0.1, ETF('SPY'): 0.6, ETF('IEF'): 0.3},
                expected=[
                    Trade(NOW, ETF('SPY'), 60 / 293.25, 293.24, 293.25),
                    Trade(NOW, ETF('IEF'), 30 / 112.68, 112.67, 112.68),
                ],
            ),
            Pair(
                # to leveraged balanced portfolio.
                allocation={Cash(): -0.8, ETF('SPY'): 1.2, ETF('IEF'): 0.6},
                expected=[
                    Trade(NOW, ETF('SPY'), 120 / 293.25, 293.24, 293.25),
                    Trade(NOW, ETF('IEF'), 60 / 112.68, 112.67, 112.68),
                ],
            ),
            Pair(
                # to short balanced portfolio.
                allocation={Cash(): 1.9, ETF('SPY'): -0.6, ETF('IEF'): -0.3},
                expected=[
                    Trade(NOW, ETF('SPY'), -60 / 293.24, 293.24, 293.25),
                    Trade(NOW, ETF('IEF'), -30 / 112.67, 112.67, 112.68),
                ],
            ),
            Pair(
                # to leveraged short balanced portfolio.
                allocation={Cash(): 2.8, ETF('SPY'): -1.2, ETF('IEF'): -0.6},
                expected=[
                    Trade(NOW, ETF('SPY'), -120 / 293.24, 293.24, 293.25),
                    Trade(NOW, ETF('IEF'), -60 / 112.67, 112.67, 112.68),
                ],
            ),
            Pair(
                # to leveraged short mixed portfolio.
                allocation={Cash(): 0.7, ETF('SPY'): 0.6, ETF('IEF'): -0.3},
                expected=[
                    Trade(NOW, ETF('SPY'), 60 / 293.25, 293.24, 293.25),
                    Trade(NOW, ETF('IEF'), -30 / 112.67, 112.67, 112.68),
                ],
            ),
        ]
    )
    def test_make_trades_from_weights_without_holdings(self, allocation, expected):
        rebalancing = Rebalancing(
            contracts=allocation.keys(),
            allocation=allocation.values(),
            absolute=True,
            measure='weight',
            time=NOW,
        )
        broker = make_broker(deposit=100)
        actual = rebalancing.make_trades(broker)
        assert actual == expected

    def test_make_trades_from_weights_with_holdings(self):
        broker = make_broker(deposit=100)
        broker._holdings_quantity = {
            Cash(): -20,
            ETF('SPY'): 80,
            ETF('IEF'): -10
        }
        rebalancing = Rebalancing(
            contracts=[Cash(), ETF('SPY'), ETF('IEF')],
            allocation=[-0.2, -0.5, 1.7],
            time=NOW,
        )
        trades_actual = rebalancing.make_trades(broker)
        nlv = broker.net_liquidation_value()
        positions_value_diff = {
            # Target position state minus current position state
            Cash(): (-0.2 * nlv) - (-20 * 1),
            ETF('SPY'): (-0.5 * nlv) - (80 * 293.24),
            ETF('IEF'): (1.7 * nlv) - (-10 * 112.68),
        }
        trades_expected = [
            Trade(NOW, ETF('SPY'), positions_value_diff[ETF('SPY')] / 293.24, 293.24, 293.25),
            Trade(NOW, ETF('IEF'), positions_value_diff[ETF('IEF')] / 112.68, 112.67, 112.68),
        ]
        assert trades_actual == trades_expected

    @pytest.mark.parametrize(
        argnames=['holdings', 'target', 'expected'],
        argvalues=[
            Imbalance(
                holdings={},
                target={Cash(): 100},
                expected=[],
            ),
            Imbalance(
                holdings={ETF('SPY'): 80},
                target={ETF('SPY'): 180},
                expected=[Trade(NOW, ETF('SPY'), 100, 293.24, 293.25)],
            ),
            Imbalance(
                holdings={ETF('SPY'): 10, ETF('IEF'): -5.5, ES(2020, 3): 0.41, VX(2020, 3): -1.2},
                target={},
                expected=[
                    Trade(NOW, ETF('SPY'), -10, 293.24, 293.25),
                    Trade(NOW, ETF('IEF'), 5.5, 112.67, 112.68),
                    Trade(NOW, ES(2020, 3), -0.41, 2941.75, 2942),
                    Trade(NOW, VX(2020, 3), 1.2, 19.20, 19.25),
                ],
            ),
            Imbalance(
                holdings={ETF('SPY'): 10, ETF('IEF'): -5.5, ES(2020, 3): 0.41, VX(2020, 3): -1.2},
                target={Cash(): 100},
                expected=[
                    Trade(NOW, ETF('SPY'), -10, 293.24, 293.25),
                    Trade(NOW, ETF('IEF'), 5.5, 112.67, 112.68),
                    Trade(NOW, ES(2020, 3), -0.41, 2941.75, 2942),
                    Trade(NOW, VX(2020, 3), 1.2, 19.20, 19.25),
                ],
            ),
            Imbalance(
                holdings={ES(2020, 3): 3},
                target={FutureChain(contracts=[ES(2020, 3)]): 10},
                expected=[
                    Trade(NOW, ES(2020, 3), 7, 2941.75, 2942),
                ],
            ),
        ]
    )
    def test_make_trades_from_nr_contracts(self, holdings, target, expected):
        rebalancing = Rebalancing(
            contracts=target.keys(),
            allocation=target.values(),
            absolute=True,
            measure='nr-contracts',
            time=NOW,
        )
        broker = make_broker(deposit=100, initial_holdings=holdings)
        actual = rebalancing.make_trades(broker)
        assert actual == expected

    def test_fractional(self):
        rebalancing = Rebalancing(
            contracts=[ETF('SPY'), ETF('IEF'), ES(2020, 3)],
            allocation=[21.8, -25.9, -1.9],
            absolute=True,
            measure='nr-contracts',
            fractional=False,
            time=NOW,
        )
        broker = make_broker(
            deposit=100,
            initial_holdings={
                ETF('SPY'): 0.2,
                ETF('IEF'): 25.7,
            },
        )
        actual = rebalancing.make_trades(broker)
        assert actual == [
            Trade(NOW, ETF('SPY'), 21, 293.24, 293.25),
            Trade(NOW, ETF('IEF'), -51, 112.67, 112.68),
            Trade(NOW, ES(2020, 3), -1, 2941.75, 2942),
        ]

    @pytest.mark.parametrize('weight', [0.049, -0.049])
    def test_margin_with_positive_weight(self, weight):
        rebalancing = Rebalancing(
            contracts=[ETF('SPY'), ETF('IEF'), ES(2020, 3)],
            allocation=[0.218, -0.259, weight],
            absolute=False,
            measure='weight',
            time=NOW,
            margin=0.05,
        )
        broker = make_broker()
        actual = rebalancing.make_trades(broker)
        expected = [
            Trade(NOW, ETF('SPY'), 743.3930093776642, 293.24, 293.25),
            Trade(NOW, ETF('IEF'), -2298.748557734978, 112.67, 112.68),
            # Note ES is missing because abs(weight) is < margin.
        ]
        assert actual == expected

    @pytest.mark.parametrize('sign', [-1, 1])
    def test_margin_with_liquidation_ignores_margin(self, sign):
        rebalancing = Rebalancing(
            time=NOW,
            margin=0.05,
        )
        broker = make_broker(initial_holdings={ES(2020, 3): 0.001 * sign})
        current_weight = broker.holdings_weights()[ES(2020, 3)]
        current_quantity = broker.holdings_quantity[ES(2020, 3)]
        assert abs(current_weight) < 0.05
        actual = rebalancing.make_trades(broker)
        expected = [
            Trade(NOW, ES(2020, 3), -current_quantity, 2941.75, 2942),
        ]
        assert actual == expected



"""This module provides the core logic to perform portfolio _rebalancing.
Holdings (cash and contracts) are held in a Brokerage account and historical
_rebalancing are keep in TrackRecord."""
from tradingenv.broker.trade import Trade
from tradingenv.broker.rebalancing import Rebalancing
from tradingenv.broker.fees import IBrokerFees, BrokerFees
from tradingenv.broker.track_record import TrackRecord
from tradingenv.contracts import AbstractContract, Cash
from tradingenv.exchange import Exchange
from typing import Dict, Union, Sequence
from datetime import datetime
from collections import defaultdict
import numpy as np
SECONDS_IN_YEAR = 365 * 24 * 60 * 60


class EndOfEpisodeError(Exception):
    """Raised when the net liquidation state of the account goes below zero.
    This exception will trigger an end of episode."""


class Context:
    def __init__(self, nlv, weights, values, nr_contracts,
                 margins):
        self.nlv = nlv
        self.weights = weights
        self.values = values
        self.nr_contracts = nr_contracts
        self.margins = margins

    def __eq__(self, other: 'Context') -> bool:
        return all(
            self.nlv == other.nlv and
            self.weights == other.weights and
            self.values == other.values and
            self.nr_contracts and other.nr_contracts and
            self.margins and other.margins
        )


class Broker:
    """This class allows to simulate a sequence of _rebalancing requests.

    Attributes
    ----------
    _holdings_margins : Dict[AbstractContract, float]
        A dictionary mapping contracts to margins deposited at the clearing
        house, e.g. when buying derivatives products such as futures.
    _holdings_quantity : Dict[AbstractContract, float]
        A dictionary mapping contracts to number of contracts held now.
    track_record : TrackRecord
        Keeps track of historical _rebalancing request.
    """

    def __init__(
        self,
        exchange: Exchange,
        base_currency: Cash = Cash(),
        deposit: float = 100.0,
        fees: IBrokerFees = BrokerFees(),
        epsilon: float = 1e-7
    ):
        """
        Parameters
        ----------
        exchange : Exchange
            An exchanged used to retrieve market prices and conditions to
            execute _rebalancing requests and trade accordingly.
        base_currency: Cash
            Base currency.
        deposit : float
            Initial deposit, 100 units of base currency by default.
        """
        # Inputs.
        self.exchange = exchange
        self.base_currency = base_currency
        self.fees = fees
        self._epsilon = epsilon

        # More attributes.
        self._holdings_margins: Dict[AbstractContract, float] = defaultdict(float)
        self._holdings_quantity: Dict[AbstractContract, float] = defaultdict(float)
        self._holdings_quantity[base_currency] = deposit
        self._initial_deposit = deposit
        self._last_accrual: Union[datetime, None] = None
        self._last_marking_to_market_price = dict()
        self.track_record = TrackRecord()
        
    @property
    def holdings_quantity(self) -> Dict[AbstractContract, float]:
        return dict(self._holdings_quantity)

    @property
    def holdings_margins(self) -> Dict[AbstractContract, float]:
        return dict(self._holdings_margins)

    def rebalance(self, rebalancing: Rebalancing):
        """Given a _rebalancing object, execute all trades needed to perform the
        _rebalancing.

        Developer notes
        ---------------
        The _rebalancing logic is slit between Rebalancing and Broker.
        Rebalancing is responsible for calculating (but not executing) the
        trading instructions (i.e. a list of trades), wheres Broker is
        responsible for anything else (i.e. pre-processing, execution and
        post-processing).
        This design allows to optionally compute the trades for a variety of
        _rebalancing objects and see what the trades would be, without
        performing any _rebalancing. Also, the separation of concerns between
        calculating the trades and execution the trades makes it easier to
        test the code.
        """
        rebalancing.profit_on_idle_cash = self.accrued_interest(rebalancing.time, True)
        rebalancing.context_pre = self.context()
        rebalancing.trades = rebalancing.make_trades(self)
        for trade in rebalancing.trades:
            self.transact(trade)
        rebalancing.context_post = self.context()
        self.track_record._checkpoint(rebalancing)

    def transact(self, trade: Trade):
        """
        Parameters
        ----------
        trade : Trade
            A trade instance to be executed.

        Notes
        -----
        Transacting in long lasting positions might lead to numerical errors in
        the order of ~1e-15. It's a design choice not to deal with such errors
        because for prices --> +inf, _holdings_quantity --> 0 < 1e-15. In other
        words, we don't deal with such numerical errors to avoid unexpected
        side effects when asset prices are relatively too high.
        https://floating-point-gui.de/basic/
        """
        # Marking to market here before transacting to avoid to update the
        # last marking to market price used to price the position before
        # overriding with the acquisition price at the end of this code.
        self.marking_to_market(trade.contract)

        # Calculate target _margin requirements. Note that trade will decrease
        # margins to be held whenever the exposure of the position is reduced.
        target_quantity = abs(self._holdings_quantity[trade.contract] + trade.quantity)
        margin_expected = (
            trade.acq_price
            * target_quantity
            * trade.contract.multiplier
            * trade.contract.margin_requirement
        )
        margin_actual = self._holdings_margins[trade.contract]
        margin_diff = margin_expected - margin_actual

        # Pay transaction costs.
        self._holdings_quantity[self.base_currency] -= trade.cost_of_commissions

        # Acquisition.
        self._holdings_quantity[self.base_currency] -= trade.cost_of_cash
        self._holdings_quantity[self.base_currency] -= margin_diff
        self._holdings_margins[trade.contract] += margin_diff
        self._holdings_quantity[trade.contract] += trade.quantity

        # https://stackoverflow.com/questions/588004/is-floating-point-math-broken
        # Small quantities different from zero will make nlv to retrieve
        # prices from the exchange of expired contracts, resulting in errors as:
        # ValueError: Missing liquidation transaction_price for VX(VXN04).
        # This will make also _holdings_quantity more clean, ie no numerical
        # errors. The drawback are bugs when the actual target quantity is
        # 1e-15, occuring when nlv is close to zero and huge asset price.
        if abs(self._holdings_quantity[trade.contract]) < self._epsilon:
            self._holdings_quantity[trade.contract] = 0.

        # Update _margin requirements. Bid-ask spread is implicitly paid
        # here and now.
        self._last_marking_to_market_price[trade.contract] = trade.acq_price
        self.marking_to_market(trade.contract)

    def marking_to_market(
        self, contract: Union[AbstractContract, Sequence[AbstractContract]] = None
    ):
        """
        Parameters
        ----------
        contract : Union[AbstractContract, Sequence[AbstractContract]]
            An optional contract or sequence of contracts to for which
            marking-to-market has to be performed. If blank, all contracts will
            be assumed to be updated.

        References
        ----------
        https://www.cmegroup.com/education/courses/introduction-to-futures/mark-to-market.html
        https://www.cmegroup.com/education/courses/introduction-to-futures/margin-know-what-is-needed.html
        """
        # NOTE: this might be a performance bottleneck.
        if contract is None:
            contracts = list(self._holdings_margins)
        else:
            contracts = [contract]
        for contract in contracts:
            # Vary the _margin to reflect gains or losses.
            if contract.margin_requirement == 0:
                continue
            quantity = self._holdings_quantity[contract]
            liq_price = self.exchange[contract].liq_price(quantity)
            if np.isnan(liq_price):
                continue
            try:
                last_price = self._last_marking_to_market_price[contract]
            except KeyError:
                continue
            price_change = liq_price - last_price
            profit = quantity * contract.multiplier * price_change
            self._holdings_margins[contract] += profit
            self._last_marking_to_market_price[contract] = liq_price

            # Withdraw excess _margin or deposit if _margin call.
            current_margin = self._holdings_margins[contract]
            target_margin = (
                liq_price * abs(quantity) * contract.multiplier * contract.margin_requirement
            )
            excess_margin = current_margin - target_margin
            self._holdings_margins[contract] -= excess_margin
            self._holdings_quantity[self.base_currency] += excess_margin
            if self._holdings_margins[contract] < 0:
                raise ValueError(
                    "Unexpected situation during sanity check. "
                    "Margin for {} is negative: {}"
                    "".format(contract, self._holdings_margins[contract])
                )

    def holdings_values(self, kind: str = "notional") -> Dict[AbstractContract, float]:
        """Returns np.array with state of all positions within the portfolio
        (cash included) in base currency. The array has as many entries as
        traded contracts.
        Notional values, so transaction_price * multiplier."""
        # Note: when quantity is zero, value is trivially zero. Skip instead?
        holdings_values = defaultdict(float)
        for contract, quantity in self._holdings_quantity.items():
            if quantity == 0:
                value = 0.0
            else:
                order_book = self.exchange[contract]
                liq_price = (
                    order_book.bid_price if quantity >= 0 else order_book.ask_price
                )
                if np.isnan(liq_price):
                    raise ValueError(
                        "Missing liquidation transaction_price for {}.".format(contract)
                    )
                if kind == "notional":
                    value = quantity * liq_price * contract.multiplier
                elif kind == "liquidation":
                    value = contract.cash_requirement * quantity * liq_price
                    value += self._holdings_margins[contract]
                else:
                    raise ValueError("Unsupported 'kind'.")
            holdings_values[contract] = value
        return holdings_values

    def holdings_weights(self) -> Dict[AbstractContract, float]:
        """Relative weights of the notional traded value over NLV.
        Returns current weights of the portfolio as positions values
        (including market spread) divided by the net liquidation state of the
        account (including market spread)."""
        nlv = self.net_liquidation_value()
        holdings_notional_values = self.holdings_values()
        return {
            contract: value / nlv
            for contract, value in holdings_notional_values.items()
        }

    def net_liquidation_value(self, raise_if_broke: bool = True) -> float:
        """
        Parameters
        ----------
        raise_if_broke : bool
            An EndOfEpisodeError will be raised if the notional liquidation
            value of the account is non-positive.

        Returns
        -------
        The net liquidation value of the account pricing long (short) position
        using bid (ask) prices.
        """
        self.marking_to_market()
        holdings_values = self.holdings_values(kind="liquidation")
        nlv = sum(holdings_values.values())
        if raise_if_broke and nlv <= 0:
            raise EndOfEpisodeError(
                "Net liquidation state of the account is ${}".format(nlv)
            )
        return nlv

    def accrued_interest(self, now: datetime, accrue: bool = False) -> float:
        """
        Parameters
        ----------
        now : datetime
            Current time, used to calculate how long it has passed since the
            last time that interest rates were deposited or charged.
        accrue : bool
            If True, the interest rate will be accrued, i.e. charged if
            negative or deposited if positive. The interest rate applied is
            determined by Broker.fees.interest_rate.

        Notes
        -----
        No interest rate paid on margins. This is a realistic conservative
        assumption (e.g. Interactive Brokers does not pay interest rates on
        margins and financial instrument cannot be deposited as _margin).

        Returns
        -------
        Amount of interest rate owed (either positive or negative) since
        when if was last deposited/charged from the account's holdings.
        """
        if self._last_accrual is None:
            self._last_accrual = now
        if now < self._last_accrual:
            raise ValueError("now={} < last_update={}".format(now, self._last_accrual))
        years = (now - self._last_accrual).total_seconds() / SECONDS_IN_YEAR
        order_book = self.exchange[self.fees.interest_rate]
        amount = self._holdings_quantity[self.base_currency]

        # Compounded annual growth rate earned (paid) on idle (borrowed) cash.
        cagr = order_book.mid_price - self.fees.markup * np.sign(amount)
        rate_period = (1 + cagr) ** years - 1
        accrued_interest = amount * rate_period
        if amount > 0. and accrued_interest < 0.:
            accrued_interest = 0.
        if accrue:
            self._holdings_quantity[self.base_currency] += accrued_interest
            self._last_accrual = now
        return accrued_interest

    def context(self) -> Context:
        return Context(
            nlv=self.net_liquidation_value(),
            weights=self.holdings_weights(),
            values=self.holdings_values(),
            nr_contracts=self.holdings_quantity,
            margins=self.holdings_margins,
        )



from tradingenv.broker.fees import IBrokerFees, BrokerFees
from tradingenv.broker.trade import Trade
from tradingenv.contracts import ETF, Rate, ES
from datetime import datetime
from abc import ABC


class TestAbstractBrokerFees:
    def test_is_abc(self):
        assert issubclass(IBrokerFees, ABC)

    def test_commissions_is_abstractmethod(self):
        assert IBrokerFees.commissions.__isabstractmethod__

    def test_initialization(self):
        class Fees(IBrokerFees):
            def commissions(self, trade):
                return 0.0

        fees = Fees(0.18)
        assert fees.markup == 0.18
        assert fees.interest_rate == Rate("FED funds rate")
        assert fees.proportional == 0.0
        assert fees.fixed == 0.0


class TestBrokerFees:
    def test_commissions(self):
        fees = BrokerFees()
        trade = Trade(datetime.now(), ETF("SPY"), 2.0, 3.0, 4.0, fees)
        assert fees.commissions(trade) == 0.0

    def test_commissions_proportional_when_negative(self):
        fees = BrokerFees(proportional=0.01)
        trade = Trade(datetime.now(), ES(2019, 6), -2.0, 3.0, 4.0, fees)
        assert fees.commissions(trade) == abs(trade.notional) * fees.proportional

    def test_commissions_fixed_when_negative(self):
        fees = BrokerFees(fixed=0.01)
        trade = Trade(datetime.now(), ES(2019, 6), -2.0, 3.0, 4.0, fees)
        assert fees.commissions(trade) == fees.fixed

    def test_commissions_proportional_and_fixed_when_negative(self):
        fees = BrokerFees(proportional=0.01, fixed=0.02)
        trade = Trade(datetime.now(), ES(2019, 6), -2.0, 3.0, 4.0, fees)
        assert (
            fees.commissions(trade)
            == fees.fixed + abs(trade.notional) * fees.proportional
        )

    def test_commissions_proportional_when_positive(self):
        fees = BrokerFees(proportional=0.01)
        trade = Trade(datetime.now(), ES(2019, 6), 2.0, 3.0, 4.0, fees)
        assert fees.commissions(trade) == abs(trade.notional) * fees.proportional

    def test_commissions_fixed_when_positive(self):
        fees = BrokerFees(fixed=0.01)
        trade = Trade(datetime.now(), ES(2019, 6), 2.0, 3.0, 4.0, fees)
        assert fees.commissions(trade) == fees.fixed

    def test_commissions_proportional_and_fixed_when_positive(self):
        fees = BrokerFees(proportional=0.01, fixed=0.02)
        trade = Trade(datetime.now(), ES(2019, 6), 2.0, 3.0, 4.0, fees)
        assert (
            fees.commissions(trade)
            == fees.fixed + abs(trade.notional) * fees.proportional
        )



"""The user is expected to instance these classes when instancing the action
observation_space within TradingEnv. For examples, see spaces.py and env.py."""
from tradingenv.events import IEvent, EventContractDiscontinued
from abc import ABC, abstractmethod
from typing import List, Type, Union, Sequence
from datetime import datetime, timedelta
from bisect import bisect_right
from collections import defaultdict
from pandas.tseries.offsets import BDay
import numpy as np
import pandas as pd
import calendar


class AbstractContract(ABC):
    """Trading contract must be represented by implementation of this abstract
    class, then passed to the action observation_space within the trading environment.

    Attributes
    ----------
    now : datetime
        Current time. This is handy for a subset of contracts such as
        FutureChain or calculate nr days to expiry for derivatives. This is a
        class attribute shared across all AbstractContract instances.
    """

    # TODO: global attribute does not reset across tests and could lead to
    #   information leakage across runs. E.g. run a regression test of env
    #   without resetting. Then run a unit test using this now attribute, it
    #   will show terminal now of the regression test despite being in a unit
    #   test. Not serious problem for normal pkg usage, but not perfect design
    #   for sure. E.g. this global attr could be a Clock class to centralise
    #   time.
    now: datetime = datetime.min

    def __hash__(self) -> int:
        """Contracts are mapped by this hash number in the Exchange. Therefore,
        different contracts have to have different hash number."""
        return hash(self.symbol)

    def __eq__(self, other: Union["AbstractContract", str]) -> bool:
        """Allows using AbstractContract.symbol in addition to the class
        instance when selecting items from a dictionary."""
        try:
            return self.symbol == other.symbol
        except AttributeError:
            # AttributeError: 'other' object has no attribute 'symbol'
            return hash(self.symbol) == hash(other)

    def __repr__(self) -> str:
        return "{}({})".format(self.__class__.__name__, self.symbol)

    def __lt__(self, other: 'Future'):
        return self.symbol < other.symbol

    @property
    @abstractmethod
    def symbol(self) -> str:
        """Unique symbol associated with the concrete implementation.
        Non-unique symbols might lead to silent bugs as symbols are used
        to hash AbstractClass instances."""

    @property
    @abstractmethod
    def multiplier(self) -> float:
        """Contract's multiplier. This is generally 1 for non-derivatives
        (e.g. ETFs, Indexes, Stocks) and > 1 for derivatives (e.g. Future)."""

    @property
    @abstractmethod
    def margin_requirement(self) -> float:
        """A number ranging from 0 to 1 defining what (constant) proportion of
        the notional value should be deposited as initial _margin when buying
        or selling the contract. Note that by design we are making the
        simplifying assumption that initial _margin == maintenance _margin.
        This is generally 1 for non-derivatives (e.g. ETFs, Indexes, Stocks)
        and > 1 for derivatives (e.g. Future)."""

    @property
    @abstractmethod
    def cash_requirement(self) -> float:
        """A number ranging from 0 to 1 defining what (constant) proportion of
        the notional value should be paid when buying the contract. This is
        generally 1 for non-derivatives (e.g. ETFs, Indexes, Stocks)
        and 0 for derivatives (e.g. Future)."""

    @property
    def underlyings(self) -> List["AbstractContract"]:
        """Returns list of all subcontracts. This is needed when implementing
        composite contract such as LeadingContract. For example for
        FutureChain(ES) this would return the list of ES contracts
        across all expiry dates. Note that this is not to be confused with
        the underlying index e.g. Index('S&P 500') for eMini S&P 500."""
        return [self]

    def verify(self, mid_price: float):
        if mid_price < 0:
            raise ValueError(
                "Found negative price {} for contract {}."
                "".format(mid_price, self)
            )

    def make_events(self) -> List[IEvent]:
        """Returns a list of events (IEvent instances) to be added the
         Transmitter when initializing TradingEnv. For example,
         EventContractDiscontinued associated with futures contracts."""
        return list()

    def size(self, price: float) -> float:
        """
        Parameters
        ----------
        price : float
            Value of one unit of the underlying asset.

        Returns
        -------
        Value of the underlying asset multiplier by the future's multiplier.
        """
        return self.multiplier * price

    def static_hashing(self) -> 'AbstractContract':
        """Subclasses of AbstractContract are not guaranteed to have static
        hashing, as it may vary with time. This method
        returns an instance of AbstractContract with the same hash of the
        class instance calling this method, which is guaranteed to be static.
        See FutureChain for more info as it is an example of AbstractContract
        with dynamic hashing."""
        return self

    @property
    def symbol_short(self) -> str:
        """Symbol is guaranteed to be unique across all contracts, and in fact
        it is used to hash the instance. The short symbol is not guaranteed
        to be unique. Useful e.g. for futures where the same future across
        different expiry dates share the same short symbol but not the same
        symbol."""
        return self.symbol


class Asset(AbstractContract):
    """An index (e.g. S&P 500).

    Examples
    --------
    >>> index = Index('S&P 500')
    >>> index
    Index(S&P 500)
    """

    multiplier = 1.0
    cash_requirement = 1.0
    margin_requirement = 0.0

    def __init__(self, symbol: str):
        self._symbol = symbol

    @property
    def symbol(self) -> str:
        return self._symbol


class Cash(Asset):
    """A cash contract (e.g. USD).

    Examples
    --------
    >>> cash = Cash('EUR')
    >>> cash
    Cash(EUR)
    """

    def __init__(self, symbol: str = "USD"):
        super().__init__(symbol)


class ETF(Asset):
    """An ETF contract (e.g. SPY).

    Examples
    --------
    >>> etf = ETF('SPY')
    >>> etf
    ETF(SPY)
    """


class Index(Asset):
    """An ETF contract (e.g. SPY).

    Examples
    --------
    >>> etf = Index('SPY')
    >>> etf
    Index(SPY)
    """


class Stock(Asset):
    """A stock (e.g. Miscrosoft).

    Examples
    --------
    >>> stock = Stock('MSFT')
    >>> stock
    Stock(MSFT)
    """


class Rate(AbstractContract):
    """An rate (e.g. FED Funds rate).

    Examples
    --------
    >>> rate = Rate('FED funds rate')
    >>> rate
    Rate(FED funds rate)
    """

    multiplier = 1.0
    cash_requirement = 1.0
    margin_requirement = 0.0

    def __init__(self, symbol: str):
        self._symbol = symbol

    @property
    def symbol(self) -> str:
        return self._symbol

    def verify(self, mid_price: float):
        if mid_price >= 0.25:
            raise ValueError(
                '{} is unexpectedly high ({:.2%}). Perhaps you forgot to '
                'divide your interest rates data by 100 (0.01 reads 1%, 1 '
                'reads 100%)?'
                ''.format(self, mid_price)
            )


class Future(AbstractContract):
    """A future contract (e.g. ES, ZN)"""
    exists_since = datetime(2000, 1, 1)
    exists_until = datetime.now().date() + timedelta(days=260 * 2)
    cash_requirement = 0.0
    month_codes = {
        1: "F",
        2: "G",
        3: "H",
        4: "J",
        5: "K",
        6: "M",
        7: "N",
        8: "Q",
        9: "U",
        10: "V",
        11: "X",
        12: "Z",
    }

    def __init__(self, year: int, month: int):
        """
        Parameters
        ----------
        year : int
            Expiry year of the future contract.
        month : int
            Expiry month of the future contract.
        """
        self.expiry = self._get_expiry_date(year, month)
        self.last_trading_date = self._get_last_trading_date(self.expiry)
        self._symbol_short = self.__class__.__name__
        self._symbol = "{symbol_short}{month_code}{year_code}".format(
            symbol_short=self._symbol_short,
            month_code=self.month_codes[self.expiry.month],
            year_code=self.expiry.strftime("%y"),
        )

    def __lt__(self, other: 'Future'):
        return self.last_trading_date < other.last_trading_date

    @abstractmethod
    def _get_expiry_date(self, year: int, month: int) -> datetime:
        """Given the expiry year (int) and month (int), this method returns
         the expiry date of the contract. Note that the expiry date does not
         necessarily corresponds to the last trading date, which generally
         precedes the expiry month. The last trading date is implemented in
         Future._get_last_trading_date."""

    @abstractmethod
    def _get_last_trading_date(self, expiry: datetime) -> datetime:
        """Given the expiry date, returns the last trading date of the
        contract. Futures contract generally stops trading at least a few
        days before the expiry."""

    @property
    @abstractmethod
    def freq(self) -> str:
        """Returns frequency of the contracts, e.g. Q-DEC for ES, M for VX."""

    @property
    @abstractmethod
    def margin_requirement(self) -> float:
        """A number ranging from 0 to 1 defining what (constant) proportion of
        the notional value should be deposited as initial _margin when buying
        or selling the contract. Note that by design we are making the
        simplifying assumption that initial _margin == maintenance _margin."""

    @property
    def symbol(self) -> str:
        """Long symbol of the future contract, e.g. 'ESU19' for the ES contract
        expiring in """
        return self._symbol

    @property
    def symbol_short(self) -> str:
        return self._symbol_short

    def make_events(self) -> List[IEvent]:
        """Returns an event run when the contract will expire thus becoming
        no longer investable."""
        return [EventContractDiscontinued(time=self.expiry, contract=self)]

    def lifespan(self) -> float:
        """Returns the number of trading days left between now and the expiry
        date."""
        return np.busday_count(self.now.date(), self.expiry.date())


class FutureChain(AbstractContract):
    """Composition of Future contracts, where the only investable contract is
    the one associated with the leading month. You must have a very good
    reason for NOT using this class when running futures trading strategies.
    Thanks to this class, you don't have to work with continuous futures data,
    you can just use the chain of uncombined futures contracts across all
    expires dates of interest."""

    cash_requirement = 0.0

    def __init__(
        self,
        future_cls: Type[Future]=None,
        start: Union[str, datetime]=None,
        end: Union[str, datetime]=None,
        contracts: Sequence[Future]=None,
        month: int=0,
    ):
        """
        Parameters
        ----------
        future_cls
            A subclass of Feature (e.g. ES, ZN, VX). Full list of available
            futures can be found in futures.py or provide your own
            implementation.
        start
            Start period of the first available futures contract, passed to
            pandas.date_range.
        end
            Last period of the first available futures contract, passed to
            pandas.date_range.
        contracts
            A sequence of Future instances which will be represented by this
            single class. If you pass this argument, there is no need to also
            pass 'future_cls', 'start' or 'end' as they will be deduced.
        month
            0 by default, meaning that the preferred feature in the term
            structure is the one with the shorted expiry date (i.e. M1). If
            1, then the second expiration date will be preferred etc.

        Examples
        --------
        >>> from tradingenv.contracts import ES
        >>> future = FutureChain(ES, '2018-03', '2018-12')
        >>> future.lead_contract(datetime(2018, 4, 1))
        ES(ESM18)
        >>> future = FutureChain(ES, '2018-03', '2018-12', month=1)
        >>> future.lead_contract(datetime(2018, 4, 1))
        ES(ESU18)
        """
        self._month = month
        if future_cls is None and contracts is None:
            raise ValueError(
                "At least one argument between 'future_cls' and 'futures_"
                "chain' must be provided."
            )
        if contracts is None:
            start = start or future_cls.exists_since
            end = end or future_cls.exists_until
            months = pd.date_range(start, end, freq=future_cls.freq)
            self.contracts = [future_cls(q.year, q.month) for q in months]
        else:
            self.contracts = sorted(contracts)
        self._last_trading_dates = [
            future.last_trading_date for future in self.contracts
        ]

    @property
    def symbol(self) -> str:
        """Long symbol of the current leading contract from the current time
        (i.e. FutureChain.now). For example, in 2019 Oct, the leading ES
        contract symbol is 'ESZ19'."""
        return self.lead_contract().symbol

    @property
    def symbol_short(self) -> str:
        return self.lead_contract().symbol_short

    @property
    def multiplier(self) -> float:
        """Multiplier of the future contract."""
        return self.contracts[0].multiplier  # all the same, same cls

    @property
    def underlyings(self) -> List["AbstractContract"]:
        return self.contracts

    @property
    def margin_requirement(self) -> float:
        return self.contracts[0].margin_requirement

    def _lead_contract_idx(self, now: datetime = None) -> int:
        """Returns the index of the leading contract from self.contracts."""
        if now is None:
            now = self.now
        idx = bisect_right(self._last_trading_dates, now)
        idx += self._month
        return idx

    def lead_contract(self, now: datetime = None, month: int=0) -> Future:
        """Use self.now to identify the lead future contract"""
        if now is None:
            now = self.now
        # TODO: test month.
        idx = self._lead_contract_idx(now)
        idx += month
        return self.contracts[idx]

    def make_events(self) -> List[IEvent]:
        """Returns list of EventContractDiscontinued associated with every
        future contract belonging to the specified period when instancing this
        class."""
        events = list()
        for future in self.contracts:
            events.extend(future.make_events())
        return events

    def lifespan(self) -> float:
        """Returns the number of trading days left between now and the expiry
        date of the leading future contract."""
        return self.lead_contract().lifespan()

    def static_hashing(self) -> 'AbstractContract':
        """The hash of this class instance varies with the lead contract, which
        varies with time. This method returns the future instance associated
        with the current time and static hashing."""
        return self.lead_contract()


class ES(Future):
    """
    Notes
    -----
    First trading date is 1997-09-07, see [1].

    References
    ----------
    [1] https://www.cmegroup.com/media-room/historical-first-trade-dates.html#equityIndices
    https://www.interactivebrokers.com/en/index.php?f=26662
    https://www.cmegroup.com/clearing/risk-management/historical-margins.html
    https://www.interactivebrokers.com/en/index.php?f=1567&p=physical
    https://www.tradingacademy.com/lessons/article/futures-contract-rollover
    """
    exists_since = datetime(1997, 9, 7)  # see reference [1]
    freq = "Q-DEC"
    multiplier = 50.0

    def _get_expiry_date(self, year: int, month: int) -> datetime:
        """Returns a datetime object representing the third Friday of the
        month."""
        # https://www.barchart.com/futures/quotes/ES*0/profile
        _, nr_days = calendar.monthrange(year, month)
        dates = defaultdict(list)
        for day in range(1, nr_days + 1):
            date = datetime(year, month, day)
            weekday = date.strftime("%A")
            dates[weekday].append(date)
        return dates["Friday"][2]

    def _get_last_trading_date(self, expiry: datetime) -> datetime:
        return expiry - timedelta(days=8)

    @property
    def margin_requirement(self) -> float:
        return 0.1


class VX(Future):
    """
    References
    ----------
    [1] https://www.cmegroup.com/media-room/historical-first-trade-dates.html#equityIndices
    https://www.interactivebrokers.com/en/index.php?f=2222&exch=cfe&showcategories=FUTGRP#productbuffer
    file:///home/federico/Downloads/Trading_VIX_Futures_Options.pdf
    http://www.cboe.com/micro/vix/pdf/VIX%20fact%20sheet%202019.pdf
    http://cfe.cboe.com/cfe-products/vx-cboe-volatility-index-vix-futures/contract-specifications
    https://www.cboe.com/ms/vix-futures-specsheet.pdf
    """
    exists_since = datetime(2004, 3, 26)  # see reference [1]
    freq = "M"
    multiplier = 1000.0

    def _get_expiry_date(self, year: int, month: int) -> datetime:
        """
        Returns
        -------
        Final Settlement Date. The final settlement date for a contract with
        the "VX" ticker symbol is on the Wednesday that is 30 days prior to
        the third Friday of the calendar month immediately following the
        month in which the contract expires. The final settlement date for
        a futures contract with the "VX" ticker symbol followed by a number
        denoting the specific week of a calendar year is on the Wednesday of
        the week specifically denoted in the ticker symbol.
        If that Wednesday or the Friday that is 30 days following that
        Wednesday is a Cboe Options holiday, the final settlement date for
        the contract shall be on the business day immediately preceding that
        Wednesday.

        Reference
        ---------
        https://www.cboe.com/tradable_products/vix/vix_futures/specifications/
        """
        this_month = datetime(year, month, 1)
        next_month = this_month + timedelta(days=32)
        next_month_third_friday_day = 21 - (
                    calendar.weekday(next_month.year, next_month.month,
                                     1) + 2) % 7
        next_month_third_friday = next_month.replace(
            day=next_month_third_friday_day)
        expiration = next_month_third_friday - timedelta(days=30)
        return expiration

    def _get_last_trading_date(self, expiry: datetime) -> datetime:
        return expiry - BDay(2)

    @property
    def margin_requirement(self) -> float:
        return 0.5


class _Treasury(Future):
    """
    References
    ----------
    [1] https://www.cmegroup.com/media-room/historical-first-trade-dates.html#equityIndices
    https://www.interactivebrokers.com/en/index.php?f=26662
    https://www.cmegroup.com/clearing/risk-management/historical-margins.html
    https://www.interactivebrokers.com/en/index.php?f=1567&p=physical
    https://www.tradingacademy.com/lessons/article/futures-contract-rollover
    https://www.barchart.com/futures/quotes/ES*0/profile
    """
    exists_since = datetime(1970, 1, 1)
    freq = "Q-DEC"

    def _get_expiry_date(self, year: int, month: int) -> datetime:
        """Last business day of the delivery month."""
        # https://www.barchart.com/futures/quotes/ZN*0/profile
        dates = pd.date_range(datetime(year, month, 1), periods=31, freq="B")
        dates = [date for date in dates if date.month == month]
        return dates[-1]

    def _get_last_trading_date(self, expiry: datetime) -> datetime:
        # It's like ES, but physical delivery. See page 18:
        # https://www.cmegroup.com/education/files/understanding-treasury-futures.pdf
        # https://www.barchart.com/futures/quotes/ZN*0/profile
        return (expiry - timedelta(days=30)).replace(day=24)

    @property
    def margin_requirement(self) -> float:
        return 0.05


class ZQ(_Treasury):
    multiplier = 4167
    margin_requirement = 0.004


class ZT(_Treasury):
    multiplier = 2000
    margin_requirement = 0.02


class ZF(_Treasury):
    multiplier = 1000
    margin_requirement = 0.02


class ZN(_Treasury):
    multiplier = 1000
    margin_requirement = 0.03


class ZB(_Treasury):
    multiplier = 1000
    margin_requirement = 0.05


class NK(Future):
    """
    References
    ----------
    https://www.cmegroup.com/trading/equity-index/international-index/nikkei-225-dollar_contract_specifications.html
    https://misc.interactivebrokers.com/cstools/contract_info/v3.10/index.php?action=Conid%20Info&wlId=IB&conid=345561869&lang=en
    https://www.cmegroup.com/trading/equity-index/nikkei-225-futures-and-options.html
    https://www.cmegroup.com/trading/equity-index/international-index/nikkei-225-dollar_contract_specifications.html
    """

    exists_since = datetime(1990, 11, 8)  # see reference [1]
    freq = "Q-DEC"
    multiplier = 5.

    def _get_expiry_date(self, year: int, month: int) -> datetime:
        """Returns a datetime object representing the third Friday of the
        month."""
        # https://www.barchart.com/futures/quotes/ES*0/profile
        _, nr_days = calendar.monthrange(year, month)
        dates = defaultdict(list)
        for day in range(1, nr_days + 1):
            date = datetime(year, month, day)
            weekday = date.strftime("%A")
            dates[weekday].append(date)
        return dates["Friday"][1]

    def _get_last_trading_date(self, expiry: datetime) -> datetime:
        """
        Termination of trading here:
            https://www.cmegroup.com/trading/equity-index/nikkei-225-futures-and-options.html
        https://misc.interactivebrokers.com/cstools/contract_info/v3.10/index.php?action=Conid%20Info&wlId=IB&conid=345561869&lang=en
        """
        # It's like ES, but physical delivery. See page 18:
        # https://www.cmegroup.com/education/files/understanding-treasury-futures.pdf
        # https://www.barchart.com/futures/quotes/ZN*0/profile
        return expiry - timedelta(days=14)  # NOTE: IB allows to trade 1 day before the expiry

    @property
    def margin_requirement(self) -> float:
        """Trading terminates on business day before the 2nd friday of the
        contract month."""
        return 0.3



import pytest
from tradingenv.features import Feature
from tradingenv.events import Observer
import gym.spaces
from datetime import datetime
from tradingenv.events import IEvent
from sklearn.preprocessing import MinMaxScaler, StandardScaler
from sklearn.exceptions import NotFittedError
import numpy as np
import pandas as pd


class EventSentiment(IEvent):
    def __init__(self, time: datetime, sentiment: float):
        self.time = time
        self.sentiment = sentiment


class TestAbstractFeature:
    def test_feature_is_an_observer(self):
        assert issubclass(Feature, Observer)

    def test_init_default_params(self):
        feature = Feature()
        assert feature.space is None
        assert feature.name == type(feature).__name__
        assert feature.transformer is None
        assert feature.save is True
        assert feature.history == dict()

    def test_init_name(self):
        feature = Feature(name='NewName')
        assert feature.name == 'NewName'

    def test_init_save(self):
        feature = Feature(save=False)
        assert feature.save is False

    def test_init_space(self):
        space = gym.spaces.Box(-100, +100, (1, ))
        feature = Feature(space)
        assert feature.space is space

    def test_call_when_parse_is_not_implemented(self):
        feature = Feature()
        assert feature() is feature

    def test_call_when_parse_is_implemented(self):
        class F(Feature):
            def parse(self):
                return 1

        space = gym.spaces.Discrete(2)
        feature = F(space)
        assert feature() == 1

    def test_call_raises_if_feature_does_not_belong_to_space(self):
        class F(Feature):
            def parse(self):
                return 5

        space = gym.spaces.Discrete(2)
        feature = F(space)
        with pytest.raises(ValueError):
            feature(verify=True)

    def test_call_does_not_verify_obs_by_default(self):
        """Here obs does not belong to space but calling feature() does not
        raise by default. This is by design for two reasons:
        (1) big features might not belong to the space until they warm up
        (2) state is verify=True by default, so avoid double verification
        """
        class F(Feature):
            def parse(self):
                return 5

        space = gym.spaces.Discrete(2)
        feature = F(space)
        # with pytest.raises(ValueError):
        feature()

    def test_call_does_not_raises_if_verify_set_to_false(self):
        class F(Feature):
            def parse(self):
                return 5

        space = gym.spaces.Discrete(2)
        feature = F(space)
        feature(verify=False)

    def test_call_raises_if_verify_set_to_true(self):
        class F(Feature):
            def parse(self):
                return 5

        space = gym.spaces.Discrete(2)
        feature = F(space)
        with pytest.raises(ValueError):
            feature(verify=True)

    def test_fit_transformer_manual_if_provided(self):
        class F(Feature):
            def _manual_fit_transformer(self):
                self.transformer.n_samples_seen_ = 0
                self.transformer.scale_ = 1.
                self.transformer.min_ = 2.
                self.transformer.data_min_ = 3.
                self.transformer.data_max_ = 4.
                self.transformer.data_range_ = 1.

        f = F(
            space=gym.spaces.Box(-1., 1., (1, 2), float),
            transformer=MinMaxScaler(),
        )
        f.fit_transformer()
        assert f.transformer.n_samples_seen_ == 0
        assert f.transformer.scale_ == 1.
        assert f.transformer.min_ == 2.
        assert f.transformer.data_min_ == 3.
        assert f.transformer.data_max_ == 4.
        assert f.transformer.data_range_ == 1.

    @pytest.mark.parametrize(
        ['shape', 'expected_shape'],
        (
                [(1, ), (10, 1)],
                [(5,), (10, 5)],
                [(1, 3), (10, 1, 3)],
                [(3, 1), (10, 3, 1)],
                [(2, 3), (10, 2, 3)],
                [(2, 3, 4), (10, 2, 3, 4)],
        )
    )
    def test_parse_history(self, shape, expected_shape):
        feature = Feature(space=gym.spaces.Box(-1., +1., shape, float))
        nr_obs = 10
        for t in pd.date_range('2022-01-01', freq='B', periods=nr_obs):
            feature._save_observation(feature.space.sample())
            feature.last_update = t
        actual = feature._parse_history()
        assert isinstance(actual, np.ndarray)
        assert actual.shape == expected_shape

    def test_fit_transformer_box_raises_if_no_history(self):
        f = Feature(
            space=gym.spaces.Box(-1., 1., (1, 2), float),
            transformer=MinMaxScaler(),
        )
        with pytest.raises(ValueError):
            f.fit_transformer()

    def test_fit_transformer_without_history(self):
        f = Feature(
            space=gym.spaces.MultiDiscrete([2, 3]),
            transformer=MinMaxScaler(),
        )
        with pytest.raises(NotImplementedError):
            f.fit_transformer()

    def test_fit_transformer_with_unsupported_space_raises(self):
        f = Feature(
            space=gym.spaces.MultiDiscrete([2, 3]),
            transformer=MinMaxScaler(),
        )
        obs = np.array([1, 1])
        f._save_observation(obs)
        with pytest.raises(NotImplementedError):
            f.fit_transformer()

    def test_reset_last_update(self):
        class F(Feature):
            def process_EventSentiment(self, event: EventSentiment):
                return event.sentiment

        feature = F()
        assert feature.last_update is None
        EventSentiment(datetime(2019, 1, 1), 0.42).notify([feature])
        assert feature.last_update == datetime(2019, 1, 1)
        feature.reset()
        assert feature.last_update is None

    def test_reset_of_custom_future_with_custom_initialisation(self):
        class F(Feature):
            def __init__(self, a, b='b'):
                super().__init__()
                self.a = a
                self.b = b

        feature = F('a')
        # Does not raise
        feature.reset()
        assert feature.a == 'a'
        assert feature.b == 'b'

    def test_save_parsed_feature_to_history_then_reset(self):
        class F(Feature):
            def process_EventSentiment(self, event: EventSentiment):
                self._last_event = event

            def parse(self):
                return self._last_event.sentiment

        feature = F()
        assert feature.history == dict()
        EventSentiment(datetime(2019, 1, 1), 0.42).notify([feature])
        assert feature.history == {datetime(2019, 1, 1): 0.42}
        feature.reset()
        assert feature.history == dict()

    def test_dont_save_history_when_save_is_false(self):
        class F(Feature):
            def process_EventSentiment(self, event: EventSentiment):
                self._last_event = event

            def parse(self):
                return self._last_event.sentiment

        feature = F(save=False)
        assert feature.history == dict()
        EventSentiment(datetime(2019, 1, 1), 0.42).notify([feature])
        assert feature.history == dict()
        feature()
        assert feature.history == dict()
        feature.reset()
        assert feature.history == dict()

    def test_a_deepcopy_is_saved_in_data(self):
        class F(Feature):
            def __init__(self):
                super().__init__()
                self.data = dict()

            def process_EventSentiment(self, event: EventSentiment):
                self.data['sentiment'] = event.sentiment

            def parse(self):
                return self.data

        feature = F()
        EventSentiment(datetime(2019, 1, 1), 0.42).notify([feature])
        EventSentiment(datetime(2019, 1, 2), 0.43).notify([feature])
        assert feature.history == {
            datetime(2019, 1, 1): {'sentiment': 0.42},
            datetime(2019, 1, 2): {'sentiment': 0.43},
        }



from tradingenv.state import IState
from tradingenv.features import Feature
import gym.spaces
import numpy as np
import unittest.mock


class TestState:
    def test_state_without_features(self):
        state = IState()
        assert state.features is None

    def test_state_with_feature_with_space_undefined(self):
        class F(Feature):
            pass

        features = [Feature()]
        state = IState(features)
        assert state.space is None

    def test_state_with_feature_with_space_defined(self):
        class F1(Feature):
            pass

        class F2(Feature):
            pass

        f1 = F1(gym.spaces.Discrete(2))
        f2 = F2(gym.spaces.Box(-1., +1., (1, ), np.float64))
        state = IState([f1, f2])
        actual = state.space
        expected = gym.spaces.Dict({f1.name: f1.space, f2.name: f2.space})
        assert actual == expected

    def test_parse_when_features_are_provided(self):
        class F1(Feature):
            def parse(self, state=None):
                return 1

        class F2(Feature):
            def parse(self, state=None):
                return np.array([0.42])

        f1 = F1(gym.spaces.Discrete(2))
        f2 = F2(gym.spaces.Box(-1., +1., (1, ), float))
        state = IState([f1, f2])
        actual = state.parse()
        expected = {
            f1.name: 1,
            f2.name: 0.42,
        }
        assert actual == expected

    def test_reset_resets_features(self):
        class F1(Feature):
            pass

        f1 = F1()
        state = IState([f1])
        with unittest.mock.patch.object(F1, 'reset', return_value=None) as reset:
            state.reset()
            reset.assert_called_once()




import tradingenv
from tradingenv.contracts import Rate
from abc import ABC, abstractmethod


class IBrokerFees(ABC):
    def __init__(
        self,
        markup: float = 0.0,
        interest_rate: Rate = Rate("FED funds rate"),
        proportional: float = 0.0,
        fixed: float = 0.0,
    ):
        """
        Parameters
        ----------
        markup : float
            A non-negative CAGR, zero by default. This mark-up is the profit
            of the broker when applying the interest rate to idle cash and
            loans. For example, if interest_rate is 0.03 (3%) and fees=0.01
            (1%), CAGR on idle cash will be 2% and cost of loans will be 4%.
        interest_rate : Rate
            This interest rate is used to: (1) generate a positive stream of
            cash flows on idle cash (e.g. when the portfolio is not fully
            invested) and (2) generate a negative stream of cash flows on
            loans (e.g. borrowed cash to pay in a leveraged portfolio of ETFs).
            By default, this is Rate('FED funds rate') and it's expressed in
            terms of compounded annual growth rate (CAGR).
        proportional : float
            Broker fees to be applied to the notional traded value of the
            security. E.g. if 0.01 (1%) mean that 1 USD is paid for every
            100 USD traded.
        fixed : float
            Fixed broker fees to be applied to each trade, regardless the
            notional traded value of trade.
        """
        self.markup = markup
        self.interest_rate = interest_rate
        self.proportional = proportional
        self.fixed = fixed

    @abstractmethod
    def commissions(self, trade: "tradingenv.broker.trade.Trade") -> float:
        """Returns total broker fees."""


class BrokerFees(IBrokerFees):
    """No broker fees of any sort will be applied when using this class."""

    def commissions(self, trade: "tradingenv.broker.trade.Trade") -> float:
        """Zero fees."""
        return self.fixed + abs(trade.notional) * self.proportional


class InteractiveBrokersFees(IBrokerFees):
    """Replicate Interactive Broker commisions.

    References
    ----------
    https://www.interactivebrokers.com/en/index.php?f=1590
    """

    def commissions(self, trade: "tradingenv.broker.trade.Trade") -> float:
        raise NotImplementedError()



from tradingenv.broker.allocation import _Allocation, NrContracts, Weights
from tradingenv.contracts import Cash, ETF, VX, ES, FutureChain
import pytest
from collections import namedtuple
_TestSubtraction = namedtuple('TestSubtraction', ['right', 'expected'])


class TestAllocation:
    def test_subclasses_dict(self):
        assert issubclass(_Allocation, dict)

    def test_initialization_with_mapping(self):
        allocation = _Allocation({ETF('SPY'): 1, ETF('IEF'): 2})
        assert allocation == {ETF('SPY'): 1, ETF('IEF'): 2}

    def test_initialization_with_keys_and_values(self):
        allocation = _Allocation(keys=[ETF('SPY'), ETF('IEF')], values=[1, 2])
        assert allocation == {ETF('SPY'): 1, ETF('IEF'): 2}

    def test_initialization_raises_if_more_values_than_keys(self):
        with pytest.raises(ValueError):
            _Allocation(keys=[ETF('SPY'), ETF('IEF')], values=[1, 2, 3])

    def test_initialization_raises_if_more_keys_than_values(self):
        with pytest.raises(ValueError):
            _Allocation(keys=[Cash(), ETF('SPY'), ETF('IEF')], values=[1, 2])

    def test_keys_have_static_hashing(self):
        future_chain = FutureChain(contracts=[ES(2020, 3), ES(2020, 6)])
        allocation = _Allocation({future_chain: 0.2})
        assert allocation == {ES(2020, 3): 0.2}
        key = list(allocation.keys())[0]
        assert not isinstance(key, FutureChain)
        assert isinstance(key, ES)

    def test_keys_whose_value_is_zero_are_dropped(self):
        allocation = _Allocation({ETF('SPY'): 1, ETF('IEF'): 0})
        assert allocation == {ETF('SPY'): 1}

    def test_keys_with_cash_are_dropped(self):
        allocation = _Allocation({ETF('SPY'): 1, Cash(): 2})
        assert allocation == {ETF('SPY'): 1}

    def test_subtraction_raises_if_different_measure(self):
        left = Weights({ETF('SPY'): -0.4})
        right = NrContracts({ETF('SPY'): -0.4})
        with pytest.raises(TypeError):
            left - right

    def test_subtraction_does_not_mutate_objects(self):
        left = _Allocation({ETF('SPY'): 1})
        right = _Allocation({ETF('SPY'): 1})
        actual = left - right
        assert id(actual) != id(left)
        assert id(actual) != id(right)
        assert left == {ETF('SPY'): 1}
        assert right == {ETF('SPY'): 1}

    @pytest.mark.parametrize('measure', ['nr-contracts', 'weight'])
    def test_subtraction_when_weights_returns_weights(self, measure):
        actual = Weights() - Weights()
        assert isinstance(actual, Weights)

    @pytest.mark.parametrize('measure', ['nr-contracts', 'weight'])
    def test_subtraction_when_nr_contracts_returns_nr_contracts(self, measure):
        actual = NrContracts() - NrContracts()
        assert isinstance(actual, NrContracts)

    @pytest.mark.parametrize(
        argnames=['right', 'expected'],
        argvalues=[
            _TestSubtraction(
                right={
                    Cash(): 0.5,
                    ETF('SPY'): -0.4,
                    ETF('IEF'): 0.1,
                    VX(2019, 3): 0.02,
                    VX(2019, 4): -0.01,
                    FutureChain(contracts=[ES(2020, 3), ES(2020, 6)]): 0.2,
                },
                expected={},
            ),
            _TestSubtraction(
                right={},
                expected={
                    Cash(): 0.5,
                    ETF('SPY'): -0.4,
                    ETF('IEF'): 0.1,
                    VX(2019, 3): 0.02,
                    VX(2019, 4): -0.01,
                    FutureChain(contracts=[ES(2020, 3), ES(2020, 6)]): 0.2,
                },
            ),
            _TestSubtraction(
                right={
                    ETF('SPY'): -0.4,
                    ETF('EEM'): 0.2,
                    VX(2019, 3): 0.015,
                },
                expected={
                    ETF('IEF'): 0.1,
                    ETF('EEM'): -0.2,
                    VX(2019, 3): 0.02 - 0.015,
                    VX(2019, 4): -0.01,
                    ES(2020, 3): 0.2,
                },
            ),
        ]
    )
    def test_subtraction(self, right, expected):
        future_chain = FutureChain(contracts=[ES(2020, 3), ES(2020, 6)])
        left = _Allocation({
            Cash(): 0.5,
            ETF('SPY'): -0.4,
            ETF('IEF'): 0.1,
            VX(2019, 3): 0.02,
            VX(2019, 4): -0.01,
            future_chain: 0.2,
        })
        actual = left - _Allocation(right)
        assert actual == _Allocation(expected)
        assert isinstance(actual, _Allocation)



from tradingenv.broker.allocation import NrContracts, Weights
from tradingenv.broker.broker import Broker
from tradingenv.events import EventNBBO
from tradingenv.exchange import Exchange
from tradingenv.contracts import Cash, ETF, VX, ES, Rate, FutureChain
from collections import namedtuple
from datetime import datetime
import pytest
Pair = namedtuple('Pair', ['allocation', 'expected'])
Imbalance = namedtuple('Pair', ['holdings', 'target', 'expected'])
NOW = datetime.now()


def make_broker(deposit: float = 1000000, initial_holdings=None):
    initial_holdings = initial_holdings or dict()
    nbbos = [
        EventNBBO(datetime.now(), Cash(), 1, 1),
        EventNBBO(datetime.now(), Rate("FED funds rate"), 0.0182, 0.0182),
        EventNBBO(datetime.now(), ETF('SPY'), 293.24, 293.25),
        EventNBBO(datetime.now(), ETF('IEF'), 112.67, 112.68),
        EventNBBO(datetime.now(), VX(2020, 3), 19.20, 19.25),
        EventNBBO(datetime.now(), ES(2020, 3), 2941.75, 2942),
        EventNBBO(datetime.now(), ES(2020, 6), 2940, 2940.25),
        EventNBBO(datetime.now(), ES(2020, 9), 2938.75, 2939),
        EventNBBO(datetime.now(), ES(2020, 12), 2935.25, 2935),
    ]
    exchange = Exchange()
    for nbbo in nbbos:
        exchange.process_EventNBBO(nbbo)
    broker = Broker(exchange=exchange, base_currency=Cash(), deposit=deposit)
    for contract, quantity in initial_holdings.items():
        broker._holdings_quantity[contract] = quantity
    return broker


class TestWeights:
    def test_to_nr_contracts_returns_nr_contracts_instance(self):
        broker = make_broker()
        weights = Weights({})
        assert isinstance(weights._to_nr_contracts(broker), NrContracts)

    def test_to_weights_returns_weights_instance(self):
        broker = make_broker()
        weights = Weights({})
        assert isinstance(weights._to_weights(broker), Weights)

    def test_to_weights_returns_copy(self):
        broker = make_broker()
        weights = Weights({})
        assert id(weights._to_weights(broker)) != id(weights)

    def test_to_weights_is_nr_contracts(self):
        """Allocation is already expressed as quantity, to self is returned."""
        broker = make_broker(
            initial_holdings={
                Cash(): -20,
                ETF('SPY'): 80,
                ETF('IEF'): -10,
                ES(2020, 9): 0.001,
            }
        )
        allocation = Weights(
            mapping={
                Cash(): -10,
                ETF('SPY'): 20,
                VX(2020, 3): -5,
                FutureChain(contracts=[ES(2020, 3), ES(2020, 6), ES(2020, 9)]): 2.2,
            },
        )
        actual = allocation._to_weights(broker)
        expected = {ETF('SPY'): 20, VX(2020, 3): -5, ES(2020, 3): 2.2}
        assert actual == expected

    def test_equality(self):
        broker = make_broker()
        weights = Weights({ETF('SPY'): 1})
        assert weights._to_weights(broker) == weights

    @pytest.mark.parametrize(
        argnames=['holdings'],
        argvalues=[
            [{Cash(): 20, ETF('SPY'): 80}],
            [{Cash(): -20, ETF('SPY'): 80}],
            [{Cash(): -20, ETF('SPY'): 80, ETF('IEF'): 10}],
            [{Cash(): -20, ETF('IEF'): 10}],
        ],
    )
    def test_to_nr_contracts_does_not_depend_on_broker_holdings(self, holdings):
        """Note that .as_quantity is not responsible for applying any filter,
        but just to convert the unit measure to whatever. Here we test only for
        what .as_quantity is responsible for."""
        broker = make_broker(initial_holdings=holdings)
        weights = Weights({ETF('SPY'): 1})
        actual = weights._to_nr_contracts(broker)
        acq_price = broker.exchange[ETF('SPY')].acq_price(1)
        expected = {ETF('SPY'): broker.net_liquidation_value() / acq_price}
        assert actual == expected

    def test_to_nr_contracts_round_trip(self):
        broker = make_broker()
        weights_start = Weights(
            mapping={
                ETF('SPY'): 0.8,
                VX(2020, 3): 0.5,
                FutureChain(contracts=[ES(2020, 3), ES(2020, 6), ES(2020, 9)]): -0.3,
            },
        )
        nr_contracts = weights_start._to_nr_contracts(broker)
        weights_end = nr_contracts._to_weights(broker)
        assert weights_start == weights_end

    @pytest.mark.parametrize(
        argnames=['allocation', 'expected'],
        argvalues=[
            Pair(
                allocation={
                    Cash(): -0.1,
                    ETF('SPY'): 0.2,
                    VX(2020, 3): -0.5,
                    FutureChain(contracts=[ES(2020, 3), ES(2020, 6), ES(2020, 9)]): 2.2,
                },
                expected={
                    ETF('SPY'): 0.6820119352088662,
                    VX(2020, 3): -0.026041666666666668,
                    ES(2020, 3): 0.01495581237253569,
                },
            ),
            Pair(
                allocation={
                    Cash(): 0.1,
                    ETF('SPY'): 0.2,
                    VX(2020, 3): -0.5,
                    FutureChain(contracts=[ES(2020, 3), ES(2020, 6), ES(2020, 9)]): 2.2,
                },
                expected={
                    ETF('SPY'): 0.6820119352088662,
                    VX(2020, 3): -0.026041666666666668,
                    ES(2020, 3): 0.01495581237253569,
                },
            ),
            Pair(
                allocation={
                    Cash(): 0.1,
                    ETF('SPY'): 0.2,
                    VX(2020, 3): 0.5,
                    FutureChain(contracts=[ES(2020, 3), ES(2020, 6), ES(2020, 9)]): 2.2,
                },
                expected={
                    ETF('SPY'): 0.6820119352088662,
                    VX(2020, 3): 0.025974025974025972,
                    ES(2020, 3): 0.01495581237253569,
                },
            ),
            Pair(
                allocation={
                    Cash(): -0.1,
                    ETF('SPY'): -0.2,
                    VX(2020, 3): -0.5,
                    FutureChain(contracts=[ES(2020, 3), ES(2020, 6), ES(2020, 9)]): -2.2,
                },
                expected={
                    ETF('SPY'): -0.6820351930159596,
                    VX(2020, 3): -0.026041666666666668,
                    ES(2020, 3): -0.014957083368743096,
                },
            ),
            Pair(
                allocation={
                    Cash(): -0.1,
                    ETF('SPY'): -0.2,
                    VX(2020, 3): 0,
                    FutureChain(contracts=[ES(2020, 3), ES(2020, 6), ES(2020, 9)]): -2.2,
                },
                expected={
                    ETF('SPY'): -0.6820351930159596,
                    ES(2020, 3): -0.014957083368743096,
                },
            ),
            Pair(
                allocation={
                    Cash(): 0,
                    ETF('SPY'): -0.2,
                    VX(2020, 3): 0,
                    FutureChain(contracts=[ES(2020, 3), ES(2020, 6), ES(2020, 9)]): -2.2,
                },
                expected={
                    ETF('SPY'): -0.6820351930159596,
                    ES(2020, 3): -0.014957083368743096,
                },
            ),
            Pair(
                allocation={
                    Cash(): 0,
                    ETF('SPY'): 0,
                    VX(2020, 3): 0,
                    FutureChain(contracts=[ES(2020, 3), ES(2020, 6), ES(2020, 9)]): 0,
                },
                expected={},
            ),
            Pair(
                allocation={
                    Cash(): 2,
                    ETF('SPY'): 0,
                    VX(2020, 3): 0,
                    FutureChain(contracts=[ES(2020, 3), ES(2020, 6), ES(2020, 9)]): 0,
                },
                expected={},
            ),
        ]
    )
    def test_to_nr_contracts_conversion(self, allocation, expected):
        """Note that .as_quantity is not responsible for applying any filter,
        but just to convert the unit measure to whatever. Here we test only for
        what .as_quantity is responsible for."""
        nlv = 1000
        broker = make_broker(deposit=nlv)
        allocation = Weights(allocation)
        actual = allocation._to_nr_contracts(broker)
        assert actual == expected

        # For documentation, it follows the logic to calculate expected.
        expected = {}
        for contract in [ETF('SPY'), VX(2020, 3), ES(2020, 3)]:
            weight = allocation.get(contract, 0)
            acq_price = broker.exchange[contract].acq_price(weight)
            multiplier = contract.multiplier
            if weight != 0:
                expected[contract] = weight * nlv / acq_price / multiplier
        assert actual == expected


class TestNrContracts:
    def test_to_nr_contracts_returns_nr_contracts_instance(self):
        broker = make_broker()
        nr_contracts = NrContracts({})
        assert isinstance(nr_contracts._to_nr_contracts(broker), NrContracts)

    def test_to_weights_returns_weights_instance(self):
        broker = make_broker()
        nr_contracts = NrContracts({})
        assert isinstance(nr_contracts._to_weights(broker), Weights)

    def test_to_nr_contracts_returns_copy(self):
        broker = make_broker()
        nr_contracts = NrContracts({})
        assert id(nr_contracts._to_nr_contracts(broker)) != id(nr_contracts)

    def test_to_nr_contracts_is_nr_contracts(self):
        """Allocation is already expressed as quantity, to self is returned."""
        broker = make_broker(
            initial_holdings={
                Cash(): -20,
                ETF('SPY'): 80,
                ETF('IEF'): -10,
                ES(2020, 9): 0.001,
            }
        )
        allocation = NrContracts(
            mapping={
                Cash(): -10,
                ETF('SPY'): 20,
                VX(2020, 3): -5,
                FutureChain(contracts=[ES(2020, 3), ES(2020, 6), ES(2020, 9)]): 2.2,
            },
        )
        actual = allocation._to_nr_contracts(broker)
        expected = {ETF('SPY'): 20, VX(2020, 3): -5, ES(2020, 3): 2.2}
        assert actual == expected

    def test_equality(self):
        broker = make_broker()
        nr_contracts = NrContracts({ETF('SPY'): 1})
        assert nr_contracts._to_nr_contracts(broker) == nr_contracts

    def test_to_weights_round_trip(self):
        broker = make_broker()
        nr_contracts_start = NrContracts(
            mapping={
                ETF('SPY'): 2,
                VX(2020, 3): 5.1,
                FutureChain(contracts=[ES(2020, 3), ES(2020, 6), ES(2020, 9)]): -0.3,
            },
        )
        weights = nr_contracts_start._to_weights(broker)
        nr_contracts_end = weights._to_nr_contracts(broker)
        assert nr_contracts_start == nr_contracts_end



__package__ = "tradingenv"
__version__ = "0.1.1"

# Modules.
import tradingenv.broker.broker
import tradingenv.policy
import tradingenv.env
import tradingenv.events
import tradingenv.exchange
import tradingenv.metrics
import tradingenv.rewards
import tradingenv.spaces
import tradingenv.transmitter
import tradingenv.contracts

# Common classes.
from tradingenv.env import TradingEnv
from tradingenv.spaces import BoxPortfolio
from tradingenv.state import IState
from tradingenv.transmitter import Transmitter
from tradingenv.broker.fees import BrokerFees
from tradingenv.policy import AbstractPolicy
from tradingenv.events import (
    IEvent,
    EventDone,
    EventNBBO,
    EventReset,
    EventStep,
    EventNewDate,
)



from tradingenv.contracts import (
    AbstractContract,
    Cash,
    ETF,
    Index,
    Rate,
    Future,
    FutureChain,
    VX,
)
from tradingenv.events import EventContractDiscontinued
from datetime import datetime, timedelta
import numpy as np
import pytest
import pandas as pd


class TestAbstractContract:
    def test_multiplier_is_abstractproperty(self):
        assert AbstractContract.multiplier.__isabstractmethod__

    def test_symbol_is_abstractproperty(self):
        assert AbstractContract.symbol.__isabstractmethod__

    def test_margin_requirement_is_abstractproperty(self):
        assert AbstractContract.margin_requirement.__isabstractmethod__

    def test_cash_requirement_is_abstractproperty(self):
        assert AbstractContract.cash_requirement.__isabstractmethod__

    def test_now_is_datetime_global_attribute(self):
        assert isinstance(AbstractContract.now, datetime)

    def test_make_events_returns_no_events_by_default(self):
        class MockAbstractContract(AbstractContract):
            cash_requirement = 1.0
            margin_requirement = 0.0

            @property
            def multiplier(self):
                return 42.0

            @property
            def symbol(self):
                return "TheAnswer"

        contract = MockAbstractContract()
        events = contract.make_events()
        assert events == list()

    def test_contract_in_dictionary_can_be_hashed_by_symbol(self):
        class ContractSPY(AbstractContract):
            cash_requirement = 1.0
            margin_requirement = 0.0

            @property
            def multiplier(self):
                return 18.0

            @property
            def symbol(self):
                return "SPY"

        class ContractIEF(AbstractContract):
            cash_requirement = 1.0
            margin_requirement = 0.0

            @property
            def multiplier(self):
                return 42.0

            @property
            def symbol(self):
                return "IEF"

        spy = ContractSPY()
        ief = ContractIEF()
        d = {spy: spy, ief: ief}
        assert d[spy] is spy
        assert d["SPY"] is spy
        assert d[ief] is ief
        assert d["IEF"] is ief

    def test_two_contracts_in_dictionary_with_same_hash(self):
        """This is equivalent to {'a': 1, 'a': 2} but with contracts as keys."""

        class ContractSPY1(AbstractContract):
            cash_requirement = 1.0
            margin_requirement = 0.0

            @property
            def multiplier(self):
                return 18.0

            @property
            def symbol(self):
                return "SPY"

        class ContractSPY2(AbstractContract):
            cash_requirement = 1.0
            margin_requirement = 0.0

            @property
            def multiplier(self):
                return 42.0

            @property
            def symbol(self):
                return "SPY"

        spy1 = ContractSPY1()
        spy2 = ContractSPY2()
        d = {spy1: spy1, spy2: spy2}
        assert d == {spy1: spy1} or d == {spy2: spy2}

    def test_size(self):
        class MockAbstractContract(AbstractContract):
            cash_requirement = 1.0
            margin_requirement = 0.0

            @property
            def multiplier(self):
                return 42.0

            @property
            def symbol(self):
                return "TheAnswer"

        contract = MockAbstractContract()
        assert contract.size(price=2845.25) == 2845.25 * 42

    def test_include(self):
        class SPY(AbstractContract):
            cash_requirement = 1.0
            margin_requirement = 0.0

            @property
            def multiplier(self):
                return 1.0

            @property
            def symbol(self):
                return "SPY"

        class IEF(AbstractContract):
            cash_requirement = 1.0
            margin_requirement = 0.0

            @property
            def multiplier(self):
                return 1.0

            @property
            def symbol(self):
                return "IEF"

        assert SPY() in [SPY()]
        assert SPY() not in [IEF()]

    def test_set(self):
        class SPY(AbstractContract):
            cash_requirement = 1.0
            margin_requirement = 0.0

            @property
            def multiplier(self):
                return 1.0

            @property
            def symbol(self):
                return "SPY"

        assert {SPY(), SPY()} == {SPY()}

    def test_underlyings(self):
        class SPY(AbstractContract):
            cash_requirement = 1.0
            margin_requirement = 0.0

            @property
            def multiplier(self):
                return 1.0

            @property
            def symbol(self):
                return "SPY"

        spy = SPY()
        assert spy.underlyings == [spy]

    def test_static_hashing(self):
        class ContractSPY(AbstractContract):
            cash_requirement = 1.0
            margin_requirement = 0.0

            @property
            def multiplier(self):
                return 18.0

            @property
            def symbol(self):
                return "SPY"

        spy = ContractSPY()
        assert id(spy.static_hashing()) == id(spy)

    def test_symbol_short(self):
        class ContractSPY(AbstractContract):
            cash_requirement = 1.0
            margin_requirement = 0.0

            @property
            def multiplier(self):
                return 18.0

            @property
            def symbol(self):
                return "SPY"

        spy = ContractSPY()
        assert spy.symbol_short == 'SPY'


class TestCash:
    def test_parent_class_is_abstract(self):
        assert issubclass(Cash, AbstractContract)

    def test_multiplier(self):
        cash = Cash()
        assert cash.multiplier == 1

    def test_symbol(self):
        cash = Cash()
        assert cash.symbol == "USD"

    def test_custom_symbol(self):
        cash = Cash("EUR")
        assert cash.symbol == "EUR"

    def test_margin_requirement(self):
        cash = Cash()
        assert cash.margin_requirement == 0

    def test_cash_requirement(self):
        cash = Cash()
        assert cash.cash_requirement == 1


class TestETF:
    def test_parent_class_is_abstract(self):
        assert issubclass(ETF, AbstractContract)

    def test_multiplier(self):
        cash = ETF("SPY")
        assert cash.multiplier == 1

    def test_symbol(self):
        cash = ETF("SPY")
        assert cash.symbol == "SPY"

    def test_margin_requirement(self):
        cash = ETF("SPY")
        assert cash.margin_requirement == 0

    def test_cash_requirement(self):
        cash = ETF("SPY")
        assert cash.cash_requirement == 1


class TestIndex:
    def test_parent_class_is_abstract(self):
        assert issubclass(Index, AbstractContract)

    def test_multiplier(self):
        cash = Index("S&P 500")
        assert cash.multiplier == 1

    def test_symbol(self):
        cash = Index("S&P 500")
        assert cash.symbol == "S&P 500"

    def test_margin_requirement(self):
        cash = Index("S&P 500")
        assert cash.margin_requirement == 0

    def test_cash_requirement(self):
        cash = Index("S&P 500")
        assert cash.cash_requirement == 1


class TestRate:
    def test_parent_class_is_abstract(self):
        assert issubclass(Rate, AbstractContract)

    def test_multiplier(self):
        cash = Rate("FED Funds Rate")
        assert cash.multiplier == 1

    def test_symbol(self):
        cash = Rate("FED Funds Rate")
        assert cash.symbol == "FED Funds Rate"

    def test_margin_requirement(self):
        cash = Rate("FED Funds Rate")
        assert cash.margin_requirement == 0

    def test_cash_requirement(self):
        cash = Rate("FED Funds Rate")
        assert cash.cash_requirement == 1


class TestFuture:
    def make_es_cls(self):
        class ES(Future):
            cash_requirement = 0.0
            margin_requirement = 0.1
            multiplier = 50.

            @classmethod
            def freq(cls) -> str:
                return "Q-DEC"

            def _get_expiry_date(self, year: int, month: int) -> datetime:
                return datetime(2019, 9, 17)

            def _get_last_trading_date(self, expiry: datetime) -> datetime:
                return datetime(2019, 9, 5)

        return ES(2019, 6)

    def make_zn_cls(self):
        class ZN(Future):
            cash_requirement = 0.0
            margin_requirement = 0.2
            multiplier = 100.

            @classmethod
            def freq(cls) -> str:
                return "Q-DEC"

            def _get_expiry_date(self, year: int, month: int) -> datetime:
                return datetime(2019, 9, 15)

            def _get_last_trading_date(self, expiry: datetime) -> datetime:
                return datetime(2019, 9, 4)

        return ZN(2019, 6)

    def test_parent_class_is_abstract(self):
        assert issubclass(Future, AbstractContract)

    def test_type_is_abstract(self):
        with pytest.raises(TypeError):
            Future(2019, 6)

    def test_freq_is_abstract(self):
        assert Future.freq.__isabstractmethod__

    def test_multiplier_is_abstract(self):
        assert Future.multiplier.__isabstractmethod__

    def test_get_expiry_date_is_abstract(self):
        assert Future._get_expiry_date.__isabstractmethod__

    def test_get_cutoff_is_abstract(self):
        assert Future._get_last_trading_date.__isabstractmethod__

    def test_margin_requirement_is_abstractproperty(self):
        assert Future.margin_requirement.__isabstractmethod__

    def test_month_codes(self):
        assert Future.month_codes == {
            1: "F",
            2: "G",
            3: "H",
            4: "J",
            5: "K",
            6: "M",
            7: "N",
            8: "Q",
            9: "U",
            10: "V",
            11: "X",
            12: "Z",
        }

    def test_expiry(self):
        future = self.make_es_cls()
        assert future.expiry == datetime(2019, 9, 17)

    def test_cutoff(self):
        future = self.make_es_cls()
        assert future.last_trading_date == datetime(2019, 9, 5)

    def test_symbol(self):
        future = self.make_es_cls()
        assert future.symbol == "ESU19"

    def test_make_events(self):
        future = self.make_es_cls()
        actual = future.make_events()
        expected = EventContractDiscontinued(
            contract=future, time=datetime(2019, 9, 17)
        )
        assert actual == [expected]

    def test_size(self):
        future = self.make_es_cls()
        assert future.size(price=2842) == 2842 * future.multiplier

    def test_short_symbol(self):
        future = self.make_es_cls()
        assert future.symbol_short == 'ES'

    def test_margin_requirement(self):
        future = self.make_es_cls()
        assert future.margin_requirement == 0.1

    def test_cash_requirement(self):
        future = self.make_es_cls()
        assert future.cash_requirement == 0.0

    def test_less_then_raises_if_contracts_of_different_types(self):
        es = self.make_es_cls()
        zn = self.make_zn_cls()
        assert zn.last_trading_date < es.last_trading_date
        assert zn < es

    def test_exists_since(self):
        assert Future.exists_since == datetime(2000, 1, 1)

    def test_exists_until(self):
        expected = datetime.now().date() + timedelta(days=260 * 2)
        assert Future.exists_until == expected

    def test_lifespan(self):
        now_default = Future.now
        es = self.make_es_cls()
        Future.now = es.expiry
        assert es.lifespan() == 0
        Future.now = es.expiry - timedelta(days=14)
        assert es.lifespan() == 10
        Future.now = now_default


class TestFutureChain:
    def make_es_cls(self):
        class ES(Future):
            multiplier = 50.0
            margin_requirement = 0.1
            freq = "Q-DEC"

            def _get_expiry_date(self, year: int, month: int) -> datetime:
                return datetime(year, month, 18)

            def _get_last_trading_date(self, expiry: datetime) -> datetime:
                return datetime(expiry.year, expiry.month, 8)

        return ES

    def test_subclasses_abstract_contract(self):
        assert issubclass(FutureChain, AbstractContract)

    def test_raise_if_minimal_args_are_not_provided(self):
        with pytest.raises(ValueError):
            FutureChain()

    def test_initialization_with_futures_chain(self):
        ES = self.make_es_cls()
        future_chain = FutureChain(
            contracts=[ES(2019, 6), ES(2019, 3), ES(2019, 9)]
        )
        assert future_chain.contracts == [ES(2019, 3), ES(2019, 6), ES(2019, 9)]

    def test_initialization_without_futures_chain_start_end_provided(self):
        ES = self.make_es_cls()
        future_chain = FutureChain(
            future_cls=ES, start=datetime(2019, 1, 1), end=datetime(2020, 1, 1)
        )
        assert future_chain.contracts == [ES(2019, 3), ES(2019, 6), ES(2019, 9), ES(2019, 12)]

    def test_initialization_without_futures_chain_start_end_not_provided(self):
        ES = self.make_es_cls()
        future_chain = FutureChain(future_cls=ES)
        expected = [ES(t.year, t.month) for t in pd.date_range(start=ES.exists_since, end=ES.exists_until, freq=ES.freq)]
        assert future_chain.contracts == expected

    def test_multiplier(self):
        ES = self.make_es_cls()
        future_chain = FutureChain(ES, "2019-01-01", "2019-12-31")
        assert future_chain.multiplier == 50

    def test_short_symbol(self):
        ES = self.make_es_cls()
        future_chain = FutureChain(ES, "2019-01-01", "2019-12-31")
        assert future_chain.symbol_short == 'ES'

    def test_symbol(self):
        ES = self.make_es_cls()
        future_chain = FutureChain(ES, "2019-01-01", "2019-12-31")
        assert future_chain.symbol == "ESH19"
        future_chain.now = datetime(2019, 11, 1)
        assert future_chain.symbol == "ESZ19"
        future_chain.now = datetime(2019, 5, 1)
        assert future_chain.symbol == "ESM19"
        future_chain.now = datetime.min
        assert future_chain.symbol == "ESH19"
    
    def test_lead_contract(self):
        ES = self.make_es_cls()
        future_chain = FutureChain(ES, "2019-01-01", "2020-12-31")
        assert future_chain.lead_contract() == ES(2019, 3)
        future_chain.now = datetime(2019, 11, 1)
        assert future_chain.lead_contract() == ES(2019, 12)
        future_chain.now = datetime(2019, 5, 1)
        assert future_chain.lead_contract() == ES(2019, 6)
        future_chain.now = datetime.min
        assert future_chain.lead_contract() == ES(2019, 3)

    def test_lead_contract_when_month_is_specified(self):
        ES = self.make_es_cls()
        future_chain = FutureChain(ES, "2019-01-01", "2020-12-31", month=1)
        assert future_chain.lead_contract() == ES(2019, 6)
        future_chain.now = datetime(2019, 11, 1)
        assert future_chain.lead_contract() == ES(2020, 3)
        future_chain.now = datetime(2019, 5, 1)
        assert future_chain.lead_contract() == ES(2019, 9)
        future_chain.now = datetime.min
        assert future_chain.lead_contract() == ES(2019, 6)

    def test_lead_contract_now_is_provided(self):
        ES = self.make_es_cls()
        future_chain = FutureChain(ES, "2019-01-01", "2019-12-31")
        assert future_chain.lead_contract(future_chain.now) == ES(2019, 3)
        assert future_chain.lead_contract(datetime(2019, 11, 1)) == ES(2019, 12)
        assert future_chain.lead_contract(datetime(2019, 5, 1)) == ES(2019, 6)
        assert future_chain.lead_contract(datetime.min) == ES(2019, 3)

    def test_make_events(self):
        ES = self.make_es_cls()
        future_chain = FutureChain(ES, "2019-01-01", "2019-12-31")
        actual = future_chain.make_events()
        expected = [
            EventContractDiscontinued(contract=ES(2019, 3), time=ES(2019, 3).expiry),
            EventContractDiscontinued(contract=ES(2019, 6), time=ES(2019, 6).expiry),
            EventContractDiscontinued(contract=ES(2019, 9), time=ES(2019, 9).expiry),
            EventContractDiscontinued(contract=ES(2019, 12), time=ES(2019, 12).expiry),
        ]
        assert actual == expected

    def test_size(self):
        ES = self.make_es_cls()
        future_chain = FutureChain(ES, "2019-01-01", "2019-12-31")
        assert future_chain.size(2847.25) == 2847.25 * ES.multiplier

    def test_underlyings(self):
        ES = self.make_es_cls()
        future_chain = FutureChain(ES, "2019-01-01", "2019-12-31")
        assert future_chain.underlyings == [
            ES(2019, month) for month in [3, 6, 9, 12]
        ]

    def test_margin_requirement(self):
        ES = self.make_es_cls()
        future_chain = FutureChain(ES, "2019-01-01", "2019-12-31")
        assert future_chain.margin_requirement == 0.1

    def test_cash_requirement(self):
        ES = self.make_es_cls()
        future_chain = FutureChain(ES, "2019-01-01", "2019-12-31")
        assert future_chain.cash_requirement == 0.0

    def test_lifespan(self):
        now_default = Future.now
        ES = self.make_es_cls()
        future_chain = FutureChain(ES, "2019-01-01", "2019-12-31")
        Future.now = future_chain.lead_contract().expiry
        assert future_chain.lifespan() == 0
        Future.now = future_chain.lead_contract().expiry - timedelta(days=14)
        assert future_chain.lifespan() == 10
        Future.now = now_default

    def test_static_hashing(self):
        now_default = Future.now
        ES = self.make_es_cls()
        future_chain = FutureChain(ES, "2019-01-01", "2019-12-31")
        assert future_chain.static_hashing() == ES(2019, 3)
        future_chain.now = datetime(2019, 7, 3)
        assert future_chain.static_hashing() == ES(2019, 9)
        Future.now = now_default


@pytest.mark.parametrize('year, month, expected', [
    (2020, 10, datetime(2020, 10, 21)),
    (2020, 11, datetime(2020, 11, 18)),
    (2020, 12, datetime(2020, 12, 16)),
    (2021, 1, datetime(2021, 1, 20)),
    (2021, 2, datetime(2021, 2, 17)),
    (2021, 3, datetime(2021, 3, 17)),
    (2021, 4, datetime(2021, 4, 21)),
    (2021, 5, datetime(2021, 5, 19)),
])
def test_vx_expiration(year, month, expected):
    # https://cdn.cboe.com/resources/aboutcboe/Cboe2020FUTURESCalendar.pdf
    # https://cdn.cboe.com/resources/aboutcboe/Cboe2021FUTURESCalendar.pdf
    vx = VX(year, month)
    assert vx.expiry == expected



"""Good things to test in new features are:
- .__init__
- .parse
- .fit_transform
"""
import gym.spaces

from tradingenv.events import EventNBBO
from tradingenv.library import FeaturePortfolioWeight, FeaturePrices, FeatureSpread, FeatureIsRTH
from tradingenv.exchange import Exchange
from tradingenv.contracts import ETF
from sklearn.preprocessing import MinMaxScaler, StandardScaler
from datetime import datetime
import pytest
import numpy as np
import pandas as pd


class TestPortfolioWeight:
    def test_init_single_contract(self):
        contract = ETF('SPY')
        feature = FeaturePortfolioWeight(contract, 0., 1.)
        assert feature.contracts == [contract]
        assert feature.space.low == 0.
        assert feature.space.high == 1.
        assert feature.space.shape == (1, 1)
        assert isinstance(feature.transformer, MinMaxScaler)
        assert feature.transformer.feature_range == (-1, 1.)

    def test_init_multiple_contracts(self):
        contracts = [ETF('SPY'), ETF('IEF')]
        feature = FeaturePortfolioWeight(contracts, 0., 1.)
        assert feature.contracts == contracts
        assert feature.space.shape == (1, 2)

    def test_init_custom(self):
        feature = FeaturePortfolioWeight(
            ETF('SPY'), 0., 1., 'NewName', MinMaxScaler((-0.5, 0.5), clip=True)
        )
        assert feature.transformer.feature_range == (-0.5, 0.5)
        assert feature.name == 'NewName'
        assert feature.transformer.clip is True

    @pytest.mark.parametrize('x, expected', [
        (0., -1.),
        (0.25, -0.5),
        (0.5, 0.),
        (0.75, 0.5),
        (1., 1.),
    ])
    def test_fit_transformer_one_contract(self, x, expected):
        """x and expected must be in the shape of feature.space.shape, so using
        nested list to respect that."""
        x = [[x]]
        expected = np.array([[expected]])
        feature = FeaturePortfolioWeight(ETF('SPY'), 0., 1., (-1, 1.))
        feature.fit_transformer()
        actual = feature.transformer.transform(x)
        np.testing.assert_equal(actual, expected)

    @pytest.mark.parametrize('x, expected', [
        [(0., 0.25), (-1, -0.5)],
        [(0.25, 0.5), (-0.5, 0.)],
        [(0.5, 0.75), (0., 0.5)],
        [(0.75, 1.), (0.5, 1.)],
    ])
    def test_fit_transformer_multiple_contracts(self, x, expected):
        """x and expected must be in the shape of feature.space.shape, so using
        nested list to respect that."""
        x = [x]
        expected = np.array([expected])
        feature = FeaturePortfolioWeight([ETF('SPY'), ETF('IEF')], 0., 1., (-1, 1.))
        feature.fit_transformer()
        actual = feature.transformer.transform(x)
        np.testing.assert_equal(actual, expected)

    def test_call_transforms_observation_if_transformer_has_been_fit(self):
        """Nice to have this test but it's redundant as __call__ is not
        implemented in FeaturePortfolioWeight and it has already been tested in
        its parent class Feature.
        """

        class MockBroker:
            def holdings_weights(self):
                return {ETF('SPY'): 0.6, ETF('IEF'): 0.4}

        broker = MockBroker()
        contract = ETF('SPY')
        feature = FeaturePortfolioWeight(contract, 0., 1., (-1, 1.))
        feature.reset(broker=broker)
        feature.fit_transformer()

        actual = feature.parse()
        expected = np.array([[0.6]])
        np.testing.assert_almost_equal(actual, expected)

        actual = feature()
        expected = np.array([[0.2]])
        np.testing.assert_almost_equal(actual, expected)


class TestFeaturePrices:
    def test_init_single_contract(self):
        contract = ETF('SPY')
        feature = FeaturePrices(contract)
        assert feature.contracts == [contract]
        assert feature.space.low == 0.
        assert feature.space.high == np.inf
        assert feature.space.shape == (1, 1)
        assert isinstance(feature.transformer, StandardScaler)

    def test_init_multiple_contracts(self):
        contracts = [ETF('SPY'), ETF('IEF')]
        feature = FeaturePrices(contracts)
        assert feature.contracts == contracts
        assert feature.space.shape == (1, 2)

    def test_parse(self):
        contract = ETF('SPY')
        feature = FeaturePrices(contract)
        exchange = Exchange()
        feature.reset(exchange=exchange)
        returns = pd.Series(
            data=np.random.normal(0.05, 0.12, 10),
            index=pd.date_range('2022-01-01', freq='B', periods=10),
        )
        spread = 0.
        prices = (1 + returns).cumprod()
        for t, p in prices.items():
            nbbo = EventNBBO(t, contract, p - spread, p + spread)
            exchange.process_EventNBBO(nbbo)
            actual = feature()
            expected = np.array([[p]])
            np.testing.assert_almost_equal(actual, expected)

    def test_transform_one_contract(self):
        contract = ETF('SPY')
        feature = FeaturePrices(contract)
        exchange = Exchange()
        feature.reset(exchange=exchange)
        returns = pd.Series(
            data=np.random.normal(0.05, 0.12, 10),
            index=pd.date_range('2022-01-01', freq='B', periods=10),
        )
        spread = 0.
        prices = (1 + returns).cumprod()

        # Full pass to create a history to be used to fit the transformer.
        for t, p in prices.items():
            nbbo = EventNBBO(t, contract, p - spread, p + spread)
            exchange.process_EventNBBO(nbbo)
            actual = feature(verify=True)

        # Fit the transformer.
        feature.fit_transformer()

        # Assessment.
        for t, p in prices.items():
            nbbo = EventNBBO(t, contract, p - spread, p + spread)
            exchange.process_EventNBBO(nbbo)
            actual = feature(verify=True)
            expected = (p - prices.mean()) / prices.std(ddof=0)
            np.testing.assert_almost_equal(actual, expected)

    def test_transform_multiple_contracts(self):
        contracts = [ETF('SPY'), ETF('IEF')]
        feature = FeaturePrices(contracts)
        exchange = Exchange()
        feature.reset(exchange=exchange)
        returns = pd.DataFrame(
            data=np.random.normal(0.05, 0.12, (10, 2)),
            index=pd.date_range('2022-01-01', freq='B', periods=10),
            columns=contracts,
        )
        spread = 0.
        prices = (1 + returns).cumprod()

        # Full pass to create a history to be used to fit the transformer.
        for t, p_series in prices.iterrows():
            for c, p in p_series.items():
                nbbo = EventNBBO(t, c, p - spread, p + spread)
                exchange.process_EventNBBO(nbbo)
            actual = feature(verify=True)

        # Fit the transformer.
        feature.fit_transformer()

        # Assessment.
        feature.reset(exchange=exchange)
        for t, p_series in prices.iterrows():
            for c, p in p_series.items():
                nbbo = EventNBBO(t, c, p - spread, p + spread)
                exchange.process_EventNBBO(nbbo)
            actual = feature(verify=True)
            expected = np.array([(p_series - prices.mean()) / prices.std(ddof=0)])
            np.testing.assert_almost_equal(actual, expected)


class TestFeatureSpread:
    def test_parse_one_contract(self):
        exchange = Exchange()
        contract = ETF('VXX')
        feature = FeatureSpread(contract)
        feature.reset(exchange=exchange)

        nbbo = EventNBBO(datetime(2022, 1, 1), contract, 15.05, 15.15)
        exchange.process_EventNBBO(nbbo)

        actual = feature.parse()
        expected = - np.array([[15.05 / 15.15 - 1]])
        np.testing.assert_almost_equal(actual, expected)

    def test_parse_multiple_contracts(self):
        exchange = Exchange()
        contracts = [ETF('SPY'), ETF('IEF')]
        feature = FeatureSpread(contracts)
        feature.reset(exchange=exchange)

        nbbo = EventNBBO(datetime(2022, 1, 1), ETF('SPY'), 439.11, 439.17)
        exchange.process_EventNBBO(nbbo)
        nbbo = EventNBBO(datetime(2022, 1, 1), ETF('IEF'), 109.13, 109.15)
        exchange.process_EventNBBO(nbbo)

        actual = feature.parse()
        expected = - np.array([[439.11 / 439.17 - 1, 109.13 / 109.15 - 1]])
        np.testing.assert_almost_equal(actual, expected)

    def test_parse_clip(self):
        exchange = Exchange()
        contract = ETF('VXX')
        feature = FeatureSpread(contract, clip=0.001)
        feature.reset(exchange=exchange)

        nbbo = EventNBBO(datetime(2022, 1, 1), contract, 15.05, 15.15)
        exchange.process_EventNBBO(nbbo)

        actual = feature.parse()
        expected = np.array([[0.001]])
        np.testing.assert_almost_equal(actual, expected)

    def test_custom_weights(self):
        class FeatureCustom(FeatureSpread):
            def _make_weights(self):
                return np.array([[0.25, 0.75]])

        exchange = Exchange()
        contracts = [ETF('SPY'), ETF('IEF')]
        feature = FeatureCustom(contracts)
        feature.reset(exchange=exchange)

        nbbo = EventNBBO(datetime(2022, 1, 1), ETF('SPY'), 439.11, 439.17)
        exchange.process_EventNBBO(nbbo)
        nbbo = EventNBBO(datetime(2022, 1, 1), ETF('IEF'), 109.13, 109.15)
        exchange.process_EventNBBO(nbbo)

        actual = feature.parse()
        expected = - np.array([[
            0.25 * (439.11 / 439.17 - 1) + 0.75 * (109.13 / 109.15 - 1)
        ]])
        np.testing.assert_almost_equal(actual, expected)

    @pytest.mark.parametrize('bid,ask,expected', [
        (15.05, 15.15, 1.),
        (15.05, 15.05, -1.),
        (15.05, 15.05 / (1 + 0.001 * 0.5), 0.),
        (15.05, 15.05 / (1 + 0.001 * 0.25), -0.5),
        (15.05, 15.05 / (1 + 0.001 * 0.75), 0.5),
    ])
    def test_fit_transform(self, bid, ask, expected):
        exchange = Exchange()
        contract = ETF('VXX')
        feature = FeatureSpread(contract, clip=0.001)
        feature.reset(exchange=exchange)
        feature.fit_transformer()

        nbbo = EventNBBO(datetime(2022, 1, 1), contract, bid, ask)
        exchange.process_EventNBBO(nbbo)
        actual = feature()
        expected = np.array([[expected]])
        np.testing.assert_almost_equal(actual, expected)


class TestFeatureIsRTH:
    def test_init(self):
        feature = FeatureIsRTH()
        assert isinstance(feature.space, gym.spaces.MultiBinary)

    @pytest.mark.parametrize('t,expected', [
        (datetime(2022, 1, 1, 13), 0),  # Saturday
        (datetime(2022, 1, 2, 13), 0),  # Sunday
        (datetime(2022, 1, 5, 8, 29), 0),
        (datetime(2022, 1, 5, 8, 31), 1),
        (datetime(2022, 1, 5, 9), 1),
        (datetime(2022, 1, 5, 14, 29), 1),
        (datetime(2022, 1, 5, 15, 16), 0),
    ])
    def test_parse(self, t, expected):
        expected = np.array([expected])
        exchange = Exchange()
        feature = FeatureIsRTH()
        feature.reset(exchange=exchange)
        nbbo = EventNBBO(pd.Timestamp(t), ETF('SPY'), 439.11, 439.17)
        exchange.process_EventNBBO(nbbo)
        actual = feature.parse()
        np.testing.assert_almost_equal(actual, expected)

    def test_fit_transform(self):
        exchange = Exchange()
        feature = FeatureIsRTH()
        feature.reset(exchange=exchange)
        feature.fit_transformer()
        t = datetime(2022, 1, 5, 9)
        nbbo = EventNBBO(pd.Timestamp(t), ETF('SPY'), 439.11, 439.17)
        exchange.process_EventNBBO(nbbo)

    @pytest.mark.parametrize('t,expected', [
        (datetime(2022, 1, 1, 13), 0.),  # Saturday
        (datetime(2022, 1, 2, 13), 0.),  # Sunday
        (datetime(2022, 1, 5, 8, 29), 0.),
        (datetime(2022, 1, 5, 8, 31), 0.00246914),
        (datetime(2022, 1, 5, 9), 0.0740741),
        (datetime(2022, 1, 5, 14, 29), 0.88641975),
        (datetime(2022, 1, 5, 15, 14), 0.9975309),
    ])
    def test_parse_rth(self, t, expected):
        expected = np.array([[expected]])
        exchange = Exchange()
        feature = FeatureIsRTH(kind='rth')
        feature.reset(exchange=exchange)
        nbbo = EventNBBO(pd.Timestamp(t), ETF('SPY'), 439.11, 439.17)
        exchange.process_EventNBBO(nbbo)
        actual = feature.parse()
        np.testing.assert_almost_equal(actual, expected)

    @pytest.mark.parametrize('t,expected', [
        (datetime(2022, 1, 1, 13), 0.3333333),  # Saturday
        (datetime(2022, 1, 2, 13), 0.70114943),  # Sunday
        (datetime(2022, 1, 3, 8, 29), 0.9997446),  # Sunday
        (datetime(2022, 1, 5, 8, 29), 0.99903382),
        (datetime(2022, 1, 5, 8, 31), 0.),
        (datetime(2022, 1, 5, 9), 0.),
        (datetime(2022, 1, 5, 14, 29), 0.),
        (datetime(2022, 1, 5, 15, 14), 0.),
    ])
    def test_parse_eth(self, t, expected):
        expected = np.array([[expected]])
        exchange = Exchange()
        feature = FeatureIsRTH(kind='eth')
        feature.reset(exchange=exchange)
        nbbo = EventNBBO(pd.Timestamp(t), ETF('SPY'), 439.11, 439.17)
        exchange.process_EventNBBO(nbbo)
        actual = feature.parse()
        np.testing.assert_almost_equal(actual, expected)




import tradingenv
from tradingenv.env import TradingEnv
from tradingenv.spaces import DiscretePortfolio, BoxPortfolio
from tradingenv.transmitter import Transmitter
from tradingenv.state import IState
from tradingenv.events import EventReset, EventStep
from tradingenv.contracts import Cash, ETF, ES, FutureChain, Index, AbstractContract
from tradingenv.features import Feature
from tradingenv.events import EventNBBO
from pandas.api.types import is_numeric_dtype
from datetime import datetime
from time import strptime
import gym.spaces
import numpy as np
import pandas as pd
import os

here = os.path.abspath(os.path.dirname(__file__))
excel_etf = pd.ExcelFile(os.path.join(here, "data", "etf.xlsx"))
excel_futures = pd.ExcelFile(os.path.join(here, "data", "futures_prices.xlsx"))
excel_futures_solutions = pd.ExcelFile(
    os.path.join(here, "data", "futures_solutions.xlsx")
)
excel_btc_cny = pd.read_csv(
    os.path.join(here, "data", "high_freq_btc_cny.csv"),
    parse_dates=True,
    index_col='Time',
)
excel_btc_cny.rename({'Price': Index('BTC/CNY')}, axis='columns', inplace=True)


class TargetDiscretePortfolio(IState):
    """Feature whose state indicate the current target weights of the
    portfolio."""
    def __init__(self):
        self.space = gym.spaces.Box(low=0, high=1, shape=(3,), dtype=np.float64)
        self.value = [0, 0, 0]

    def process_EventReset(self, event: "EventReset"):
        self.value = [1, 0, 0]

    def process_EventStep(self, event: "EventStep"):
        rebalancing = event.track_record[-1]
        weight1 = 0
        weight2 = rebalancing.allocation.get('S&P 500', 0)
        weight3 = rebalancing.allocation.get('T-bills', 0)
        self.value = [weight1, weight2, weight3]

    def parse(self):
        return np.array(self.value)


class TestSimpleExamples:
    def test_minimal_example(self):
        # Sample prices parsed in AbstractTransmitter.
        prices = pd.DataFrame(
            data={
                ETF("S&P 500"): [100, 101, 103, 106],
                ETF("T-bills"): [1000, 1000.5, 1001, 1002],
            },
            index=[
                datetime(2019, 1, 1, 16),  # 4pm, close transaction_price
                datetime(2019, 1, 2, 16),  # 4pm, close transaction_price
                datetime(2019, 1, 3, 16),  # 4pm, close transaction_price
                datetime(2019, 1, 4, 16),  # 4pm, close transaction_price
            ],
        )
        transmitter = Transmitter(timesteps=prices.index)
        transmitter.add_prices(prices)

        # Instance the environment.
        action_space = DiscretePortfolio(
            contracts=[Cash(), ETF("S&P 500"), ETF("T-bills")],
            allocations=[[0, 1, 0], [0, 0.5, 0.5], [0, 0, 1]],
        )
        env = TradingEnv(
            action_space=action_space,
            state=TargetDiscretePortfolio(),
            reward="RewardSimpleReturn",
            transmitter=transmitter,
        )

        # Reset the environment.
        state = env.reset()
        np.testing.assert_equal(
            # Initial portfolio is 100% in cash.
            actual=state,
            desired=np.array([1.0, 0.0, 0.0]),
        )
        assert env._now == datetime(2019, 1, 1, 16)
        assert env.exchange[ETF("S&P 500")].mid_price == 100
        assert env.exchange[ETF("T-bills")].mid_price == 1000

        # Interact with the environment.
        state, reward, done, info = env.step(action=0)
        assert env._now == datetime(2019, 1, 2, 16)
        assert env.exchange[ETF("S&P 500")].mid_price == 101
        assert env.exchange[ETF("T-bills")].mid_price == 1000.5
        np.testing.assert_equal(
            actual=state,
            desired=np.array([0.0, 1.0, 0.0])
        )
        np.testing.assert_almost_equal(actual=reward, desired=101 / 100 - 1)
        assert not done

        # Interact with the environment.
        state, reward, done, info = env.step(action=1)
        assert env._now == datetime(2019, 1, 3, 16)
        assert env.exchange[ETF("S&P 500")].mid_price == 103
        assert env.exchange[ETF("T-bills")].mid_price == 1001
        np.testing.assert_equal(
            actual=state,
            desired=np.array([0.0, 0.5, 0.5])
        )
        np.testing.assert_almost_equal(
            actual=reward, desired=0.5 * (103 / 101 - 1) + 0.5 * (1001 / 1000.5 - 1)
        )
        assert not done

        # Interact with the environment.
        state, reward, done, info = env.step(action=2)
        assert env._now == datetime(2019, 1, 4, 16)
        assert env.exchange[ETF("S&P 500")].mid_price == 106
        assert env.exchange[ETF("T-bills")].mid_price == 1002
        np.testing.assert_equal(
            actual=state,
            desired=np.array([0.0, 0.0, 1])
        )
        np.testing.assert_almost_equal(actual=reward, desired=1002 / 1001 - 1)
        assert done

        # Verify track record.
        assert env.broker.track_record._time == [
            datetime(2019, 1, 1, 16),  # 4pm, close transaction_price
            datetime(2019, 1, 2, 16),  # 4pm, close transaction_price
            datetime(2019, 1, 3, 16),  # 4pm, close transaction_price
        ]
        nlvs_actual = list(env.broker.track_record.net_liquidation_value(before_rebalancing=False).squeeze())
        assert nlvs_actual == [100.0, 101.0, 102.02523738130934]


class TestETF:
    def get_env(self, cost_of_spread: float = 0, margin: float=0):
        prices = excel_etf.parse("Prices", index_col="Date", parse_dates=True)
        prices.columns = [ETF(str(symbol)) for symbol in prices.columns]
        transmitter = Transmitter(timesteps=prices.index)
        transmitter.add_prices(prices, spread=cost_of_spread)
        action_space = BoxPortfolio([Cash(), ETF("ETF:SPY"), ETF("ETF:IEF")], margin=margin)
        return TradingEnv(action_space=action_space, transmitter=transmitter)

    def assert_is_equal_nlv(self, actual, expected):
        assert isinstance(actual, pd.DataFrame)
        assert is_numeric_dtype(actual.squeeze())
        assert actual.index.equals(expected.index)
        assert np.isclose(actual.values, expected.values).all()

    def test_cash_only(self):
        env = self.get_env()

        # Interact with the environment.
        env.reset()
        done = False
        while not done:
            action = np.array([1.0, 0, 0])
            obs, reward, done, info = env.step(action)

        # Assert equivalence.
        actual = env.broker.track_record.net_liquidation_value(before_rebalancing=False)
        expected = excel_etf.parse("TestCashOnly", index_col="Date", parse_dates=True)[
            ["NLV"]
        ]
        # Drop duplicate time index.
        expected = expected.iloc[1:, :]
        self.assert_is_equal_nlv(actual, expected)

    def test_balanced_portfolio_rebalanced_daily(self):
        env = self.get_env()

        # Interact with the environment.
        env.reset()
        done = False
        while not done:
            action = np.array([0, 0.6, 0.4])
            obs, reward, done, info = env.step(action)

        # Assert equivalence.
        actual = env.broker.track_record.net_liquidation_value(before_rebalancing=False)
        expected = excel_etf.parse(
            sheet_name="TestDailyRebalancing6040", index_col="Date", parse_dates=True
        )
        # Drop duplicate time index.
        expected = expected.iloc[1:, :]
        self.assert_is_equal_nlv(actual, expected[["NLV"]])

    def test_buy_and_hold(self):
        env = self.get_env()

        # Interact with the environment.
        env.reset()
        done = False
        action = np.array([0, 0.6, 0.4])
        obs, reward, done, info = env.step(action)
        env.action_space._margin = np.inf
        while not done:
            obs, reward, done, info = env.step(action)

        # Assert equivalence.
        actual = env.broker.track_record.net_liquidation_value(before_rebalancing=False)
        expected = excel_etf.parse(
            sheet_name="TestBuyAndHold6040", index_col="Date", parse_dates=True
        )
        # Drop duplicate time index.
        expected = expected.iloc[1:, :]
        self.assert_is_equal_nlv(actual, expected[["NLV"]])

    def test_reproducible_simulation_after_reset(self):
        env = self.get_env()

        # Interact with the environment: first simulation.
        env.reset()
        done = False
        while not done:
            action = np.array([0, 0.6, 0.4])
            obs, reward, done, info = env.step(action)
        track_record_1 = env.broker.track_record

        # Interact with the environment: second simulation.
        env.reset()
        done = False
        while not done:
            action = np.array([0, 0.6, 0.4])
            obs, reward, done, info = env.step(action)
        track_record_2 = env.broker.track_record

        # Assert equivalence.
        nlv1 = track_record_1.net_liquidation_value(before_rebalancing=False)
        nlv2 = track_record_2.net_liquidation_value(before_rebalancing=False)
        assert nlv1.equals(nlv2)


class TestFutures:
    def test_example_runs_without_raising_exceptions(self):
        def mapper(name):
            year = int(name[27:31])
            month = strptime(name[23:26], "%b").tm_mon
            return ES(year, month)

        # Instance the env.
        prices = excel_futures.parse(
            sheet_name="ES", index_col="Date", parse_dates=True
        )
        prices.rename(mapper=mapper, axis="columns", inplace=True)
        transmitter = Transmitter(
            timesteps=prices.index, folds={"training-set": [datetime.min, datetime.max]}
        )
        transmitter.add_prices(prices)
        leading_future = FutureChain(ES, "1997-12", "2019-10")
        env = TradingEnv(action_space=[Cash(), leading_future], transmitter=transmitter)
        expected = 1 - ES(2019, 6).margin_requirement
        err = 1e-7

        # Interact with the env.
        env.reset()
        done = False
        while not done:
            action = np.array([0, 1])
            state, reward, done, info = env.step(action)

            # Margin requirement is 0.1 of notional ES value, so there should
            # be a constant 90% of cash which will generate a positive interest
            # if FED funds rate > 0.
            holdings_weights = env.broker.holdings_weights()
            cash_weight = holdings_weights[env.broker.base_currency]
            assert expected - err < cash_weight < expected + err

        # Asset.
        nlv_actual = env.broker.track_record.net_liquidation_value(before_rebalancing=False)

        # Reference of expected: S&P Futures Return in chart page 6.
        # https://www.cmegroup.com/education/files/deconstructing-futures-returns-the-role-of-roll-yield.pdf
        nlv_expected = excel_futures_solutions.parse(
            sheet_name="BuyAndHoldLeadingES", index_col="Date", parse_dates=True
        )
        nlv_expected.rename({'nlv': 'Net liquidation value'}, axis='columns', inplace=True)
        # Drop duplicate initial index.
        pd.testing.assert_frame_equal(nlv_actual, nlv_expected)


class TestHighFrequencyData:
    def test_example(self):
        env = TradingEnv(
            action_space=[Cash(), Index('BTC/CNY')],
            prices=excel_btc_cny,
        )

        env.reset()
        done = False
        while not done:
            action = np.array([0, 1])
            state, reward, done, info = env.step(action)

        nlv = env.broker.track_record.net_liquidation_value(before_rebalancing=False)
        # Discard last price from excel_btc_cny because we don't take action
        # at the last price, because there
        assert nlv.index.equals(excel_btc_cny.index[:-1])
        assert nlv.nr_observations() == len(excel_btc_cny) - 1
        np.testing.assert_approx_equal(
            actual=float((nlv.iloc[-1] / nlv.iloc[0]).item()),
            # isn't a next one to evaluate the action.
            desired=float((excel_btc_cny.iloc[-2] / excel_btc_cny.iloc[0]).item()),
        )
        nlv.tearsheet()  # does not raise


class SpotPrice(Feature):
    def __init__(self, fit_transformer: bool = False):
        super().__init__()
        self.data = dict()

    def process_EventNBBO(self, event: EventNBBO):
        if event.contract in [ETF("ETF:SPY"), ETF("ETF:IEF")]:
            self.data[event.contract] = event.mid_price

    def parse(self, state = None):
        return self.data


class TestStateAndFeatureHistory:
    """In this class we test IState and AbstractFeature.history when a list
    of features is passed to the state."""
    def test_features_and_state_history(self):
        # Note: this could be moved to a faster integration test, but kept here
        # to show an example of using state and feature history.
        # excel_etf = pd.ExcelFile(os.path.join("/home/federico/Desktop/repos/trading-gym/tests/regression/data", "etf.xlsx"))
        feature = SpotPrice()
        prices = excel_etf.parse("Prices", index_col="Date", parse_dates=True)
        prices.columns = [ETF(str(symbol)) for symbol in prices.columns]
        transmitter = Transmitter(timesteps=prices.index)
        transmitter.add_prices(prices)
        action_space = BoxPortfolio([Cash(), ETF("ETF:SPY"), ETF("ETF:IEF")])
        env = TradingEnv(
            action_space=action_space,
            transmitter=transmitter,
            state=IState([feature])
        )
        env.backtest()

        # Features' history.
        expected = prices.T.to_dict()
        actual = feature.history
        assert actual == expected

        # State history.
        expected = {k: {'SpotPrice': v} for k, v in prices.T.to_dict().items()}
        actual = env.state.history
        assert actual == expected

        # tear down, otherwise other tests relying on this global attr may fail
        # (yes, this logic needs to be improved but not trivial).
        AbstractContract.now = datetime.min

    def test_features_history_when_episode_length_is_provided(self):
        # Note: this could be moved to a faster integration test.
        feature = SpotPrice()
        prices = excel_etf.parse("Prices", index_col="Date", parse_dates=True)
        prices.columns = [ETF(str(symbol)) for symbol in prices.columns]
        transmitter = Transmitter(timesteps=prices.index)
        transmitter.add_prices(prices)
        action_space = BoxPortfolio([Cash(), ETF("ETF:SPY"), ETF("ETF:IEF")])
        env = TradingEnv(
            action_space=action_space,
            transmitter=transmitter,
            state=IState([feature])
        )
        env.backtest(episode_length=10)

        # Features' history.
        expected = prices.loc[:env.now()].T.to_dict()
        actual = feature.history
        assert actual == expected

        # State history.
        expected = {k: {'SpotPrice': v} for k, v in prices.loc[env._transmitter._start_date:env.now()].T.to_dict().items()}
        actual = env.state.history
        assert actual == expected

        # tear down, otherwise other tests relying on this global attr may fail
        # (yes, this logic needs to be improved but not trivial).
        AbstractContract.now = datetime.min



"""Note: all transformers expect 2D arrays in feature.transformer.transform.
    ValueError: Expected 2D array, got 1D array instead
Therefore, all space shapes should really be 2D and parsed history 3D. However,
            'vx-term-structure-basis': mid_prices[1:] - self.exchange[Contracts.vix].mid_price,
            'vx-term-structure-roll-yield': np.array([term_structure.roll_yield(0, t) for t in lifespan[:2]]),
            'vx-roll-yield-30days': roll_yield_30d,
            'vx-implied-sharpe-30days': implied_sharpe,
            'time-since-vix-update': np.array([hours_since_vix_update]),
"""
from tradingenv.features import Feature
from tradingenv.contracts import AbstractContract
from sklearn.preprocessing import MinMaxScaler, StandardScaler
from typing import Union, Sequence
import gym.spaces
import numpy as np
import exchange_calendars as xcals


def listify(item):
    """If the item is not iterable, insert the item in a list."""
    try:
        item = list(item)
    except TypeError:  # contracts is a single contract, not in an iterable
        item = [item]
    return item


class FeaturePortfolioWeight(Feature):
    """Portfolio weights of holdings."""

    def __init__(
            self,
            contracts: Union[AbstractContract, Sequence[AbstractContract]],
            low: float,
            high: float,
            name: str = None,
            transformer=None,
            total: bool = False,  # TODO: test
    ):
        """
        Parameters
        ----------
        contracts
            A contract or sequence of contracts.
        low
            Minimum possible portfolio allocation to the contract. E.g. -1.5
            for -150% (i.e. 150% short).
        high
            Minimum possible portfolio allocation to the contract. E.g. -1.5
            for -150% (i.e. 150% short).
        name
            Feature name. Default is class name if omitted.
        transformer
            A class from sklearn.preprocessing.

        Notes
        -----
        Because gym.spaces.Box forces the same low and high bounds on all
        items, you'll have to instance a different PortfolioWeight for each
        contract with different 'low' and 'high' allocation.
        """
        if transformer is None:
            transformer = MinMaxScaler((-1., 1.))
        self.contracts = listify(contracts)
        self.total = total
        self._size = 1 if total else len(self.contracts)
        super().__init__(
            space=gym.spaces.Box(low, high, (1, self._size), float),
            name=name,
            transformer=transformer,
        )

    def parse(self):
        """Returns array of currently held weights of self.contracts in the
        portfolio."""
        holdings = self.broker.holdings_weights()
        w = [float(holdings.get(contract, 0.)) for contract in self.contracts]
        if self.total:
            w = [sum(w)]
        return np.array([w])

    def _manual_fit_transformer(self):
        """Whatever the max or min allocation is per asset, either long or
        short, rescale the current portfolio weights in the [-1, +1] range."""
        x = np.concatenate([self.space.high, self.space.low])
        self.transformer.fit(x)


class FeaturePrices(Feature):
    """Feature representing price data."""

    def __init__(
            self,
            contracts: Union[AbstractContract, Sequence[AbstractContract]],
            low: float = 0.,
            high: float = np.inf,
            name: str = None,
            transformer=None,
    ):
        """
        Parameters
        ----------
        contracts
            A contract or sequence of contracts.
        name
            Feature name. Default is class name if omitted.
        transformer
            A class from sklearn.preprocessing.
        """
        if transformer is None:
            transformer = StandardScaler()
        self.contracts = listify(contracts)
        super().__init__(
            space=gym.spaces.Box(low, high, (1, len(self.contracts)), float),
            name=name,
            transformer=transformer,
        )

    def parse(self):
        """Parse sequence of prices."""
        w = [self.exchange[contract].mid_price for contract in self.contracts]
        return np.array([w])


class FeatureSpread(Feature):
    """Feature representing prices spreads."""

    def __init__(
            self,
            contracts: Union[AbstractContract, Sequence[AbstractContract]],
            clip: float = 0.01,
            name: str = None,
            transformer=None,
            size: int = None
    ):
        """
        Parameters
        ----------
        contracts
            A contract or sequence of contracts.
        clip
            Spread (%) greater than this are clipped. By default this is
            0.01 (i.e. 1%).
        name
            Feature name. Default is class name if omitted.
        transformer
            A class from sklearn.preprocessing. Note that spread can be highly
            positively skewed unless clipped or log transformed if the market
            data include overnight sessions, illiquid or emerging assets, in
            which case you might want to log the spread or use a more
            sophisticated transformer to normalise the data.
        """
        if transformer is None:
            transformer = MinMaxScaler(feature_range=(-1, 1))
        self.contracts = listify(contracts)
        self._low = 0.
        self._high = clip
        self._size = size or len(self.contracts)
        super().__init__(
            space=gym.spaces.Box(self._low, self._high, (1, self._size), float),
            name=name,
            transformer=transformer,
        )

    def _make_weights(self):
        """Returned array must be of length len(self.contracts).
        For example, if you want to equal weights the spread of two contracts
        this method should return np.array([[0.5, 0.5]])"""
        return np.identity(self._size)
        # lifespan = np.array([c.lifespan() for c in self.contracts])
        # vxm0_weight = min(1, ((252 / 12) - lifespan[1]) / (lifespan[0] - lifespan[1]))
        # vxm1_weight = 1 - vxm0_weight
        # return np.array([vxm0_weight, vxm1_weight])

    def parse(self):
        """Parse sequence of spreads."""
        weights = self._make_weights()
        spreads = [
            self.exchange[c].bid_price / self.exchange[c].ask_price - 1
            for c in self.contracts
        ]
        spreads = np.abs([spreads])
        spreads = spreads.dot(weights.T)
        return spreads.clip(self._low, self._high)

    def _manual_fit_transformer(self):
        """Whatever the max or min allocation is per asset, either long or
        short, rescale the current portfolio weights in the [-1, +1] range."""
        x = np.concatenate([self.space.high, self.space.low])
        self.transformer.fit(x)


class _NullTransformer:
    def fit(self, *args, **kwargs):
        return self

    def transform(self, x):
        return x


class FeatureIsRTH(Feature):
    """Boolean flag indicating weather markets is trading during Regular
    Trading Hours."""

    def __init__(
            self,
            calendar: str = 'XCBF',
            tz: str = 'America/Chicago',
            name: str = None,
            transformer=None,
            kind: str = 'binary',
    ):
        if transformer is None:
            transformer = _NullTransformer()
            self._tz = tz
        if kind == 'binary':
            space = gym.spaces.MultiBinary(1)
        else:
            space = gym.spaces.Box(0., 1., (1, 1), float)
        self.calendar = xcals.get_calendar(calendar, side='neither')
        self.kind = kind
        super().__init__(space=space, name=name, transformer=transformer)

    def parse(self):
        """Parse progress bar during RTH, ETH or just a dummy indicating
        weather """
        now = self._now().tz_localize(self._tz).tz_convert('UTC')
        is_rth = self.calendar.is_open_on_minute(now)
        if self.kind == 'rth':
            left = self.calendar.previous_open(now)
            right = self.calendar.next_close(now)
            progress = (now - left) / (right - left)
            progress = np.array([[progress]]) * int(is_rth)
            #progress = np.array([1 - progress], dtype=np.float64) * int(is_rth)
        elif self.kind == 'eth':
            left = self.calendar.previous_close(now)
            right = self.calendar.next_open(now)
            progress = (now - left) / (right - left)
            progress = np.array([[progress]]) * int(not is_rth)
            #progress = np.array([1 - progress], dtype=np.float64) * int(not is_rth)
        else:
            progress = np.array([int(is_rth)])
        return progress




from tradingenv.broker.rebalancing import Rebalancing
from tradingenv.broker.broker import Broker, Exchange
from tradingenv.events import EventNBBO
from tradingenv.contracts import Cash, Rate, ETF
from typing import List
import pandas as pd
from datetime import datetime
import pytest


class TestTrackRecord:
    def make_exchange(self, nbbo_events: List[EventNBBO] = None):
        if nbbo_events is None:
            nbbo_events = list()
        nbbo_events += [
            EventNBBO(datetime.now(), Cash(), 1, 1),
            EventNBBO(datetime.now(), Rate("FED funds rate"), 0.0, 0.0),
            EventNBBO(datetime.now(), ETF("SPY"), 145, 146),
            EventNBBO(datetime.now(), ETF("IEF"), 45, 46),
        ]
        exchange = Exchange()
        for nbbo in nbbo_events:
            exchange.process_EventNBBO(nbbo)
        return exchange

    def make_broker(self):
        return Broker(exchange=self.make_exchange())

    def test_initialization(self):
        broker = self.make_broker()
        assert broker.track_record._time == list()
        assert broker.track_record._rebalancing == dict()
        assert broker.track_record._nr_steps_to_burn == 0

    def test_checkpoint(self):
        broker = self.make_broker()
        rebalancing1 = Rebalancing(time=datetime(2019, 1, 1))
        broker.rebalance(rebalancing1)
        assert broker.track_record._time == [datetime(2019, 1, 1)]
        assert broker.track_record._rebalancing == {datetime(2019, 1, 1): rebalancing1}

        rebalancing2 = Rebalancing(time=datetime(2019, 1, 2))
        broker.rebalance(rebalancing2)
        assert broker.track_record._time == [datetime(2019, 1, 1), datetime(2019, 1, 2)]
        assert broker.track_record._rebalancing == {
            datetime(2019, 1, 1): rebalancing1,
            datetime(2019, 1, 2): rebalancing2,
        }

    def test_checkpoint_counts_burn(self):
        broker = self.make_broker()
        broker.track_record.burn = True
        assert broker.track_record._nr_steps_to_burn == 0

        # Burn is increased because the first _rebalancing trades nothing.
        rebalancing = Rebalancing(
            time=datetime(2019, 1, 1),
            contracts=[Cash(), ETF('SPY')],
            allocation=[1, 0],
        )
        broker.rebalance(rebalancing)
        assert broker.track_record._nr_steps_to_burn == 1

        # Burn is not increased because the second _rebalancing trades something.
        rebalancing = Rebalancing(
            time=datetime(2019, 1, 2),
            contracts=[Cash(), ETF('SPY')],
            allocation=[0.9, 0.1],
        )
        broker.rebalance(rebalancing)
        assert broker.track_record._nr_steps_to_burn == 1

        # Burn is not increased because we have already started trading at the
        # previous step.
        rebalancing = Rebalancing(
                time=datetime(2019, 1, 3),
                contracts=[Cash(), ETF('SPY')],
                allocation=[1, 0],
            )
        broker.rebalance(rebalancing)
        assert broker.track_record._nr_steps_to_burn == 1

    def test_checkpoint_raises_if_duplicate_time(self):
        broker = self.make_broker()
        rebalancing = Rebalancing(time=datetime(2019, 1, 1))
        broker.rebalance(rebalancing)
        with pytest.raises(ValueError):
            broker.rebalance(rebalancing)

    def test_checkpoint_converts_pandas_timestamp_to_datetime(self):
        broker = self.make_broker()
        rebalancing = Rebalancing(time=pd.Timestamp(2019, 1, 1))
        broker.rebalance(rebalancing)
        assert broker.track_record._time == [datetime(2019, 1, 1)]
        assert broker.track_record._rebalancing == {datetime(2019, 1, 1): rebalancing}

    def test_getitem_when_index_is_int(self):
        broker = self.make_broker()
        for i in range(10):
            rebalancing = Rebalancing(time=datetime(2019, 1, i + 1))
            broker.rebalance(rebalancing)
        for i in range(10):
            time = datetime(2019, 1, i + 1)
            rebalancing = broker.track_record[i]
            assert rebalancing.time == time

    def test_getitem_when_index_is_time(self):
        exchange = self.make_exchange()
        broker = Broker(exchange)
        for i in range(10):
            rebalancing = Rebalancing(time=datetime(2019, 1, i + 1))
            broker.rebalance(rebalancing)
        for i in range(10):
            time = datetime(2019, 1, i + 1)
            rebalancing = broker.track_record[time]
            assert rebalancing.time == time

    def test_len(self):
        exchange = self.make_exchange()
        broker = Broker(exchange)
        for i in range(1, 11):
            rebalancing = Rebalancing(time=datetime(2019, 1, i))
            broker.rebalance(rebalancing)
            assert len(broker.track_record) == i



from tradingenv.broker.allocation import Weights, NrContracts
import tradingenv
from tradingenv.contracts import AbstractContract
from tradingenv.broker.trade import Trade
from typing import Sequence, List
from datetime import datetime


class Rebalancing:
    """Private class returned by PortfolioSpace._make_rebalancing_request and
    passed to RebalancingResponse during its initialization."""

    def __init__(
        self,
        contracts: Sequence[AbstractContract] = None,
        allocation: Sequence[float] = None,
        measure: str = 'weight',
        absolute: bool = True,
        fractional: bool = True,
        margin: float = 0.0,
        time: datetime = None,
    ):
        """
        Parameters
        ----------
        contracts
            A sequence of contracts. The i-th contract is associated with the
            i-th element in 'allocation'.
        allocation
            A sequence of target portfolio allocations of the _rebalancing.
            The i-th allocation is associated with the i-th element
            in 'contracts'. The representation of the values in the sequence
            depends on the parameter 'kind'.
        measure
            The unit of measurement of the parameter 'allocation'. If:
            - 'weight': a 0.2 would mean 20% of the net liquidation value of
            the current portfolio.
            - 'nr-contracts': a 21 would mean 21 contracts.
        absolute
            If True (default), 'allocation' is assumed to represents the
            desired target allocation. For example, if
            allocation={ETF(SPY): 0.03} and measure='weight', then
            the desired _rebalancing will result in an allocation in ETF(SPY)
            corresponding to 3% of the net liquidation value and the remaining
            97% in cash.
            If False, 'allocation' is assumed to represent the desired change
            from the current allocation. or example, if
            allocation={ETF(SPY): 0.03} and measure='weight', then
            the desired _rebalancing will result in the previously held
            portfolio with 3% less of cash and 3% more of ETF(SPY).
        fractional
            True by default. If false, decimals will be ignored from every
            element of 'allocation'.
        margin
            A non-negative float. Rebalancing of positions which are not at
            least different by '_margin' in absolute value, will not be
            executed. Setting a small positive value (e.g. 0.02==2%) might be
            a good practice to reduce transaction costs.
            Skip trade if the absolute weight imbalance is below the margin
            AND the target weight is different from zero. The latter condition
            allows to liquidate a position in the situation where we hold in
            the portfolio a contract with a tiny weight below the margin.
        time
            Time of the _rebalancing request. Current time by default.

        Notes
        -----
        Target weights are applied to the net liquidation value of the account
        before paying transaction costs to make the problem more trackable).
        Therefore, in presence transaction costs, the amount of cash in the
        broker account will be -x where x are the broker commissions paid.
        """
        if measure == 'weight':
            self.allocation = Weights(keys=contracts, values=allocation)
        elif measure == 'nr-contracts':
            self.allocation = NrContracts(keys=contracts, values=allocation)
        else:
            raise ValueError("Unsupported argument for 'measure'.")
        self.absolute = absolute
        self.fractional = fractional
        self.margin = margin
        self.time = time or datetime.now()

        # Attributes filled when running Broker.rebalance.
        self.profit_on_idle_cash: float = ...
        self.context_pre: "tradingenv.broker.broker.Context" = ...
        self.trades: List[Trade] = ...
        self.context_post: "tradingenv.broker.broker.Context" = ...

    def make_trades(self, broker: "tradingenv.broker.Broker") -> List[Trade]:
        """
        Parameters
        ----------
        broker
            A broker instance that will be used e.g. to calculate the offset
            between target _rebalancing and current holdings.

        Returns
        -------
        A list of trade objects, recommended to perform the portfolio
        _rebalancing.
        """
        imbalance = self.allocation._to_nr_contracts(broker)
        if self.absolute:
            imbalance -= NrContracts(broker.holdings_quantity)
        weights = imbalance._to_weights(broker)
        trades = list()
        for contract, quantity in imbalance.items():
            if not self.fractional:
                # Fractional shares are not supported. Round to smallest digit.
                quantity = int(quantity)
            if abs(weights[contract]) < self.margin and contract in self.allocation:
                # Imbalance weight is smaller than margin. Skip to save costs.
                continue
            trade = Trade(
                time=self.time,
                contract=contract,
                quantity=quantity,
                bid_price=broker.exchange[contract].bid_price,
                ask_price=broker.exchange[contract].ask_price,
                broker_fees=broker.fees,
            )
            trades.append(trade)
        return trades

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self.time)
